"""
Flow Models for the HM0/1 classes
by V. Clasen, A. P. Rinaldi, F. Ciardo (2024)
"""

import numpy as np
from numba import njit
from scipy.optimize import differential_evolution
from scipy.special import erfc, expi

########################################################################################################################
########################################## ANALYTICAL FLOW MODEL #######################################################
########################################################################################################################


def Analytical_LineSource_Solution(InjectionStrategy, t_grid, r_grid, h, nu, S, k):
    """
    Args:
        InjectionStrategy (2xNt array): doublets of each consecutive step:
                                        duration of step and injection rate
                                        (only Neumann's b.c. at the well)
        t_grid (Numpy vector): consecutive time moments
        r_grid (Numpy vector): radii
        nu (scalar)          : viscosity [Pa s]
        k (scalar)           : permeability [m]
        S (scalar)           : storage (porosity * compressibility) [Pa^-1]
    Returns:
        the analytical solution stimulation factor (Numpy vector) for the
        input time moments and radii
    """
    Nt_end = len(t_grid)
    sol = np.zeros([Nt_end, len(r_grid)])
    Qsteps = np.array(InjectionStrategy[:, 1])
    Qsteps[1:] = np.diff(Qsteps)
    TimeSteps = np.array(InjectionStrategy[:, 0])
    NumOfSteps = len(Qsteps)

    for step in range(NumOfSteps):
        JJ = np.where(t_grid > TimeSteps[step])
        sol[JJ, :] = np.add(sol[JJ, :],
                            # StepInjection_Green_Function_linesource(r_grid, (t_grid[JJ] - TimeSteps[step]), Qsteps[step],
                            #                              Diff))
                            StepInjection_Green_Function_linesource(r_grid, (t_grid[JJ] - TimeSteps[step]), Qsteps[step],
                                                                    h, nu, S, k))
    return sol


def Analytical_PointSource_Solution(InjectionStrategy, t_grid, r_grid, A, nu, S, k):
    """
    Args:
        InjectionStrategy (2xNt array): doublets of each consecutive step:
                                        duration of step and injection rate
                                        (only Neumann's b.c. at the well)
        t_grid (Numpy vector): consecutive time moments
        r_grid (Numpy vector): radii
        k (scalar)           : permeability [m]
        S (scalar)           : storage (porosity * compressibility) [Pa^-1]
    Returns:
        the analytical solution stimulation factor (Numpy vector) for the
        input time moments and radii
    """
    Nt_end = len(t_grid)
    sol = np.zeros([Nt_end, len(r_grid)])
    Qsteps = np.array(InjectionStrategy[:, 1])
    Qsteps[1:] = np.diff(Qsteps)
    TimeSteps = np.array(InjectionStrategy[:, 0])
    NumOfSteps = len(Qsteps)

    for step in range(NumOfSteps):
        JJ = np.where(t_grid > TimeSteps[step])
        sol[JJ, :] = np.add(sol[JJ, :],
                            # StepInjection_Green_Function_linesource(r_grid, (t_grid[JJ] - TimeSteps[step]), Qsteps[step],
                            #                              Diff))
                            StepInjection_Green_Function_pointsource(r_grid, (t_grid[JJ] - TimeSteps[step]), Qsteps[step],
                                                                     A, nu, S, k))
    return sol


def StepInjection_Green_Function_linesource(r, t, Q, h, nu, S, k):
    # def StepInjection_Green_Function_linesource(r, t, Q, Diff, val):
    """
    Args:
        r (Numpy vector): radius of point(s) in meters
        t (Numpy vector): time moment(s) in seconds
        Q (scalar)      : source term [m^3/(s)]
        nu (scalar)     : viscosity [Pa s]
        k (scalar)      : permeability [m]
        S (scalar)      : storage (porosity * compressibility) [Pa^-1]
    Returns:
        the 1D analytical solution (Numpy vector) for continuous line source type of injection
        REF: Wu and Pruess, 2000, IJRMMS, 37, 51-61 -> Theis solution
        similar with Diff REF: pag. 261, Eq. (2) - book: “Conduction of Heat in Solids” di Carslaw and Jaeger
    """

    X = expi(-np.matmul(np.reciprocal(t.reshape((-1, 1))),
             np.square(r.reshape((1, -1)))) * nu * S / k / 4.)

    return (-Q * nu / (4. * np.pi * k * h)) * X


def StepInjection_Green_Function_pointsource(r, t, Q, A, nu, S, k):
    # def StepInjection_Green_Function_linesource(r, t, Q, Diff, val):
    """
    Args:
        r (Numpy vector): radius of point(s) in meters
        t (Numpy vector): time moment(s) in seconds
        Q (scalar)      : source term [m^3/(s)]
        nu (scalar)     : viscosity [Pa s]
        k (scalar)      : permeability [m]
        S (scalar)      : storage (porosity * compressibility) [Pa^-1]
    Returns:
        the 1D analytical solution (Numpy vector) for continuous point source type of injection
        REF: Wu and Pruess, 2000, IJRMMS, 37, 51-61 -> Theis solution
    """

    xi = np.matmul(np.reciprocal(np.sqrt(t.reshape((-1, 1)))),
                   r.reshape((1, -1))) / (np.sqrt(4. * k / S / nu))

    X = 1. / (np.sqrt(np.pi) * xi) * np.exp(-xi**2) - erfc(xi)

    return (Q * nu * r.reshape((1, -1)) / (k * A)) * X


class FlowModel_an1:
    """
        POINT SOURCE MODEL

        Initialization with radius, numpy array with time and flow, as well as
        crossectional area, model parameters initial guess, and their bounds when used for fitting
        Model parameters are the storage coefficient (namely porosity times compressibility)
        and the intrinsic rock permeability

        Simple usage for forward simulation:

        radius = np.geomspace(0.1, 2000, 30, endpoint=True) # radius for calculation
        time_Q= np.linspace(0, 1e8, 2000) # time vector for flow
        time_out = np.linspace(0, 1e8, 2000) # time vector for output, this can be different than time_Q as long as is larger
        Q = np.ones((2000)*0.01 # volumetric injection rate in m3/s
        QQ = np.vstack((time_Q.T,Q.T))
        S = 1.e-9 # storage
        k = 1.e-14 # permeability
        h = 100 #layer thickness
        rw = 0.15 #radius well
        A = 2 * np.pi * h * rw # cross sectional area
        Model = FlowModel1(radius,A,QQ,iniguess=(S, k))
        Pressure = Model.predict() # numpy vector of pressure solution at time and radius

        Simple usage for inversion:

        radius = np.geomspace(0.1, 2000, 30, endpoint=True) # radius for calculation
        TrainingTarget = 2xN array with the time moment and the pressure in Pa
        time_Q= np.linspace(0, 1e8, 2000) # time vector for flow
        Q = np.ones((2000)*0.01 # volumetric injection rate in m3/s
        TrainingData = np.vstack((time_Q.T,Q.T))
        h = 100 #layer thickness
        rw = 0.15 #radius well
        A = 2 * np.pi * h * rw # cross sectional area
        Model = FlowModel1(radius,A,TrainingData,bounds=[(1e-15 1e-5), (1e-20, 1e-12)],iniguess=(1e-9, 1e-14))
        Model.fit(TrainingTarget) # numpy vector of pressure solution at time and radius
        Pressure = Model.predict()

        For forecasting future pressure after fitting an injection Plan is needed
        Q2 = Q*2
        time2 = np.linspace(0, 1e8, 2000) +time[-1]
        Plan = np.vstack((time2.T,Q2.T)
        Prediction = Model.predict(Plan)
        """

    def __init__(self,
                 Q,  # TrainingData for fitting
                 r_grid=np.geomspace(0.1, 2000, 30, endpoint=True),
                 A=90,
                 nu=1.e-3,
                 bounds=None,
                 iniguess=(1e-9, 1e-14)):
        if bounds is None:
            bounds = [(1e-15, 1e-5), (1e-20, 1e-12)]
        self.r = r_grid
        self.A = A
        self.nu = nu
        self.Q = Q
        # self.bounds = np.log10(bounds)
        self.bounds = bounds
        self.params = iniguess
        self.model = Analytical_PointSource_Solution

    def Cost(self, x):
        S = np.power(10., x[0])
        k = np.power(10., x[1])
        _prediction = self.model(
            self.Q, self.t_target, self.r_target, self.A, self.nu, S, k) + self.observation[0]
        # _prediction = self.model(self.Q, self.t_target, self.r_target, self.A, S, k)
        Coeff_R2 = np.sum(np.square(_prediction / 1.e6
                          - self.observation / 1.e6)) / len(self.observation)

        return Coeff_R2

    def fit(self, TrainingTarget,
            popsize=10,
            polish=True,
            disp=True,
            maxiter=100,
            seed=22,
            tol=1e-5):
        """
        Args:
            TrainingTarget(Numpy array): 2xM array with the time moment and pressure measurement
        Returns:

        """
        self.t_target = TrainingTarget[:, 0]
        self.observation = TrainingTarget[:, 1:]
        self.r_target = self.r[:len(self.observation[0, :])]

        self.iter_results = []
        result = differential_evolution(self.Cost, self.bounds, popsize=popsize, polish=polish,
                                        disp=disp, maxiter=maxiter, seed=seed, tol=tol, callback=self.fm_iteration)
        self.params = np.power(10., result.x[0]), \
            np.power(10., result.x[1])

        self.hidden_mechanism = self.predict()

    def fm_iteration(self, xk, convergence):
        self.iter_results.append(xk)

    def predict(self, Plan=None, ):
        """
        Args:
            Plan (Numpy array) : 2xNf array with the time moment and the injection rate expected from that moment.
        Returns:
            (Numpy array) : 2x(Nf+N) array with the time moment and the pressure due to the plan and the training data
        """
        if Plan is None:
            Plan = []
        Q = self.Q
        t_grid = Q[:, 0]
        if (len(Plan) > 0):
            Q = np.vstack((Q, Plan))
            t_grid = Plan[:, 0]

        Prediction = self.model(Q, t_grid, self.r, self.A, self.nu,
                                self.params[0], self.params[1]) + self.observation[0]
        return Prediction


class FlowModel_an2:
    """
    LINE SOURCE MODEL (Theis' solution)

    Initialization with radius, numpy array with time and flow, as well as
    reservoir thickness, model parameters initial guess, and their bounds when used for fitting
    Model parameters are the storage coefficient (namely porosity times compressibility)
    and the intrinsic rock permeability

    Simple usage for forward simulation:

    radius = np.geomspace(0.1, 2000, 30, endpoint=True) # radius for calculation
    time_Q= np.linspace(0, 1e8, 2000) # time vector for flow
    time_out = np.linspace(0, 1e8, 2000) # time vector for output, this can be different than time_Q as long as is larger
    Q = np.ones((2000)*0.01 # volumetric injection rate in m3/s
    QQ = np.vstack((time_Q.T,Q.T))
    S = 1.e-9 # storage
    nu = 1.e-3 # viscosity
    k = 1.e-14 # permeability
    h = 100 #layer thickness
    Model = FlowModel1(radius,h,QQ,iniguess=(S, k))
    Pressure = Model.predict() # numpy vector of pressure solution at time and radius

    Simple usage for inversion:

    radius = np.geomspace(0.1, 2000, 30, endpoint=True) # radius for calculation
    TrainingTarget = 2xN array with the time moment and the pressure in Pa
    time_Q= np.linspace(0, 1e8, 2000) # time vector for flow
    Q = np.ones((2000)*0.01 # volumetric injection rate in m3/s
    TrainingData = np.vstack((time_Q.T,Q.T))
    h = 100 #layer thickness
    Model = FlowModel1(radius,h,TrainingData,bounds=[(1e-15 1e-5), (1e-20, 1e-12)],iniguess=(1e-9, 1e-14))
    Model.fit(TrainingTarget) # numpy vector of pressure solution at time and radius
    Pressure = Model.predict()

    For forecasting future pressure after fitting an injection Plan is needed
    Q2 = Q*2
    time2 = np.linspace(0, 1e8, 2000) +time[-1]
    Plan = np.vstack((time2.T,Q2.T)
    Prediction = Model.predict(Plan)
    """

    def __init__(self,
                 Q,
                 r_grid=np.geomspace(0.1, 2000, 30, endpoint=True),
                 h=100,
                 nu=1.e-3,
                 bounds=None,
                 iniguess=(1e-9, 1e-14)):
        if bounds is None:
            bounds = [(1e-15, 1e-5), (1e-20, 1e-12)]
        self.r = r_grid
        self.h = h
        self.nu = nu
        self.Q = Q

        # self.bounds = np.log10(bounds)
        self.bounds = bounds
        self.params = iniguess
        self.model = Analytical_LineSource_Solution
        # for callback
        self.best_x = None
        self.minf = np.inf

    def save_iteration_step(self, x, convergence):
        """
        function to save history of parameters and residual during DEA
        """
        # check if equal
        np.testing.assert_equal(x, self.best_x)
        self.dea_vals.append([np.power(10., x[0]),
                              np.power(10., x[1])])
        self.dea_func.append(self.minf)
        return

    def Cost(self, x):
        S = np.power(10., x[0])
        k = np.power(10., x[1])
        _prediction = self.model(
            self.Q, self.t_target, self.r_target, self.h, self.nu, S, k) + self.observation[0]
        # _prediction = self.model(self.Q, self.t_target, self.r_target, self.h, S, k)
        Coeff_R2 = np.sum(np.square(_prediction / 1.e6
                          - self.observation / 1.e6)) / len(self.observation)

        if Coeff_R2 < self.minf:
            self.minf = Coeff_R2
            self.best_x = x

        return Coeff_R2

    def fit(self, TrainingTarget,
            popsize=10,
            polish=True,
            disp=True,
            maxiter=100,
            seed=22,
            tol=1e-5):
        """
        Args:
            TrainingTarget(Numpy array): 2xM array with the time moment and pressure measurement
        Returns:

        """
        self.t_target = TrainingTarget[:, 0]
        # gets rid of the column in index nr. 1 (so the second column)
        self.observation = TrainingTarget[:, 1:]
        self.r_target = self.r[:len(self.observation[0, :])]

        # open list to save the progress of the minimization
        self.dea_vals = []
        self.dea_func = []

        result = differential_evolution(self.Cost, self.bounds, popsize=popsize, polish=polish,
                                        disp=disp, maxiter=maxiter, seed=seed, tol=tol, callback=self.save_iteration_step)

        self.params = np.power(10., result.x[0]), \
            np.power(10., result.x[1])

        self.hidden_mechanism = self.predict()

    def predict(self, Plan=None, ):
        """
        Args:
            Plan (Numpy array) : 2xNf array with the time moment and the injection rate expected from that moment.
        Returns:
            (Numpy array) : 2x(Nf+N) array with the time moment and the pressure due to the plan and the training data
        """
        if Plan is None:
            Plan = []
        Q = self.Q
        t_grid = Q[:, 0]
        if (len(Plan) > 0):
            Q = np.vstack((Q, Plan))
            t_grid = Plan[:, 0]
        Prediction = self.model(Q, t_grid, self.r, self.h, self.nu,
                                self.params[0], self.params[1]) + self.observation[0]
        return Prediction


########################################################################################################################
##################################### NUMERICAL FLOW MODEL #############################################################
########################################################################################################################
"""
Solving the RADIAL diffusion equation using the implicit finite difference (FD) approach with backward difference in 
time and second order central difference in space.
"""


def create_con_matrix(n_el):
    """
    builds up the connectivity matrix
    """
    con_mat = np.zeros((n_el, 2))
    for m in range(n_el):
        con_mat[m, 0] = m
        con_mat[m, 1] = m + 1
    return con_mat


def times_step_func(i, t):
    """
    creates the time steps in case the time stepping is not the same
    """
    return t[i + 1] - t[i]


def create_radius(r, dr):
    """
    creates the radius taking the radial position and half the distance to the next radial position
    """
    return r + 0.5 * dr


def tridiag(nx, diag, diag_above, diag_below):
    L = np.zeros((nx, nx))
    for k in range(nx - 1):
        L[k, k] = diag[k]
        L[k, k + 1] = diag_above[k]
        L[k + 1, k] = diag_below[k]
    L[nx - 1, nx - 1] = diag[nx - 1]
    return L


def Nonuniform_Tridiag(nr, con_mat, D, dt, dr_array, r):
    """
    -> solves for the matrix entries for the 1D linear pressure diffusion solution
    """
    L = np.zeros((nr, nr))
    V = np.zeros((nr, nr))
    for n in range(nr):
        elements = np.where(con_mat == n)[0]
        nod_rep = []
        el = []
        for e in elements:
            el.append(int(e))
            nodes = con_mat[e, :]
            for l in nodes:
                if not l == n:
                    idx = np.where(nodes == l)
                    nod_rep.append(int(nodes[idx]))
        if len(el) == 1:
            el.append(int(el[0]))
        nod = sorted(list(set(nod_rep)))
        beta_md = (2 * D[n] * dt) / (dr_array[el[0]] * dr_array[el[1]])
        L[n, n] = beta_md
        gamma_md = 0
        V[n, n] = gamma_md
        if len(nod) > 1:
            beta_ld = (2 * D[nod[0]] * dt) / (dr_array[el[0]]
                                              * (dr_array[el[0]] + dr_array[el[1]]))
            L[n, nod[0]] = - beta_ld
            gamma_ld = (D[nod[0]] * dt) / \
                (r[n] * (dr_array[el[0]] + dr_array[el[1]]))
            V[n, nod[0]] = gamma_ld
            beta_ud = (2 * D[nod[-1]] * dt) / (dr_array[el[1]]
                                               * (dr_array[el[0]] + dr_array[el[1]]))
            L[n, nod[-1]] = - beta_ud
            gamma_ud = (D[nod[-1]] * dt) / \
                (r[n] * (dr_array[el[0]] + dr_array[el[1]]))
            V[n, nod[-1]] = - gamma_ud
        else:
            beta_l_u_d = (2 * D[nod[0]] * dt) / (dr_array[el[0]]**2)
            L[n, nod[0]] = - beta_l_u_d
        del el, nod_rep
    V[-2, -1] = 0
    V[-1, -2] = 0
    V[-1, -1] = 0
    I = np.identity(nr)
    return I + L + V


def Numerical_Functions_Kf_change(sigma_n, pn, K_n, ap_0, ap_res, fr_s, dt, S, nu, dr_array, nr, q_f, con_mat, r):
    """
    -> solves for the pressure solution for one timestep for the radial problem including changes in permeability
    """
    q_inj = np.zeros(nr)
    sigma_n_eff = sigma_n - pn
    enc_rev = sigma_n_eff / (K_n + sigma_n_eff / (ap_0 - ap_res))
    wh = ap_0 - enc_rev
    kf = (wh**3) / (12 * fr_s)
    kf_max = (ap_0**3) / (12 * fr_s)
    if np.any(sigma_n_eff < 0.):
        idx = np.where(sigma_n_eff < 0.)
        kf[idx] = kf_max
    q_inj[0] = q_f / (wh[0] * np.min(r)**2 * np.pi * S)
    b = q_inj * dt + pn
    D = kf / (S * nu)
    L = Nonuniform_Tridiag(nr, con_mat, D, dt, dr_array, r)
    pres = np.linalg.solve(L, b)
    return pres


def Numerical_RadialSolution_Kf_invariable(S, kf, q_f, dt, nu, dr_array, nr, n_el, h, r, nt):
    """
    -> final pressure solution for all timesteps for radial problem excluding changes in permeability
    """
    q_inj = np.zeros(nr)
    p0 = np.zeros(nr)
    p_num = np.zeros((nt, nr))
    kf = np.full_like(q_inj, kf)
    pn = p0
    con_mat = create_con_matrix(n_el)
    for i in range(nt):
        q_inj[0] = q_f[i] / (h * np.min(r)**2 * np.pi * S)
        b = q_inj * dt + pn
        D = kf / (S * nu)
        L = Nonuniform_Tridiag(nr, con_mat, D, dt, dr_array, r)
        p_num_step = np.linalg.solve(L, b)
        p_num[i] = p_num_step
        pn = p_num_step
    return p_num


def Numerical_RadialSolution_Kf_variable(ap_0, ap_res, K_n, S, sigma_n, fr_s, dt, nu, dr_array, nr, n_el, q_f, r, nt):
    """
    -> final pressure solution for all timesteps for radial problem including changes in permeability
    """
    p0 = np.zeros(nr)
    p_num = np.zeros((nt, nr))
    pn = p0
    con_mat = create_con_matrix(n_el)
    for i in range(nt):
        p_num_step = Numerical_Functions_Kf_change(sigma_n, pn, K_n, ap_0, ap_res, fr_s, dt, S, nu, dr_array,
                                                   nr, q_f[i], con_mat, r)
        p_num[i] = p_num_step
        pn = p_num_step
    return p_num


class FlowModel_num:
    """
    ADD DESCRIPTION: same as FlowModel3, but without permeability changes (read below)
    """

    def __init__(self,
                 Q,
                 r_grid,
                 h,
                 params_fix,
                 bounds=None):
        if bounds is None:
            bounds = [(1e-15, 1e-5), (1e-20, 1e-12)]
        self.r = r_grid
        self.h = h
        self.params_fix = params_fix
        self.Q = Q

        # self.bounds = np.log10(bounds)
        self.bounds = bounds
        self.params = (params_fix['storage'],
                       params_fix['permeability'])
        self.model = Numerical_RadialSolution_Kf_invariable

    def translate_training(self, TrainingData, TrainingTarget, params_fix):

        self.observation = TrainingTarget[:, 1]
        self.Q = TrainingData[:, :2]

        Q = TrainingData[:, 1]
        t_grid = TrainingData[:, 0]
        self.t_grid = t_grid
        dt = abs(t_grid[0] - t_grid[1])
        nt = len(t_grid)
        r_grid = self.r  # node positions (radial)
        dr_array = np.diff(r_grid)  # spacing between nodes
        nr = len(r_grid)  # number of nodes
        n_el = len(dr_array)  # number of elements

        nu = params_fix['viscosity']

        self.args = (Q, dt, nu, dr_array, nr, n_el, self.h, r_grid, nt)

        return

    def Cost(self, x, Q, dt, nu, dr_array, nr, n_el, h, r, nt):
        S = np.power(10., x[0])
        kf = np.power(10., x[1])
        p_num = self.model(S, kf, Q, dt, nu, dr_array, nr, n_el, h, r, nt)
        p_pred = p_num[:, 0] + self.observation[0]
        Coeff = np.sum(np.square(p_pred / 1.e6
                       - self.observation / 1.e6)) / len(self.observation)

        return Coeff

    def fit(self, TrainingData, TrainingTarget):

        self.translate_training(TrainingData, TrainingTarget, self.params_fix)

        self.result = differential_evolution(func=self.Cost, bounds=self.bounds, args=self.args,
                                             maxiter=100, popsize=10, polish=True, disp=True)
        self.params = np.power(10., self.result.x[0]), \
            np.power(10., self.result.x[1])

        self.hidden_mechanism = self.predict()

    def predict(self, Plan=None, ):
        if Plan is None:
            Plan = []
        Q = self.Q
        t_grid = Q[:, 0]
        if (len(Plan) > 0):
            Q = np.vstack((Q, Plan))
            # Q = Plan[:,1]
            t_grid = Q[:, 0]
            self.args = list(self.args)
            self.args[1] = Q[:, 1]
            self.args[2] = abs(t_grid[0] - t_grid[1])
            self.args[-1] = len(t_grid)
        # r_grid = self.r
        pres_num = self.model(self.params[0], self.params[1], self.args[0], self.args[1], self.args[2], self.args[3],
                              self.args[4], self.args[5], self.args[6], self.args[7], self.args[8])
        if len(self.Q) < len(pres_num):
            Prediction = pres_num[len(self.Q):, :] + self.observation[0]
        else:
            Prediction = pres_num + self.observation[0]
        return Prediction


class FlowModel_kchangeFV:
    """
    ADD DESCRIPTION
    """

    def __init__(self,
                 Q,
                 r_grid,
                 h,
                 params_fix,
                 bounds=None):
        if bounds is None:
            bounds = [(1e-5, 1e-2), (1e-9, 1e-5), (8, 14), (-11, -7)]
        self.r = r_grid
        self.h = h
        self.params_fix = params_fix
        self.Q = Q

        # self.bounds = np.log10(bounds)
        self.bounds = bounds
        self.params = (params_fix['zero_stress_aperture'],
                       params_fix['residual_aperture'],
                       params_fix['frac_normal_stiffness'],
                       params_fix['storage'])
        self.model = Numerical_RadialSolution_Kf_variable

    def translate_training(self, TrainingData, TrainingTarget, params_fix):

        self.observation = TrainingTarget[:, 1]
        self.Q = TrainingData[:, :2]

        Q = TrainingData[:, 1]
        t_grid = TrainingData[:, 0]
        self.t_grid = t_grid
        dt = abs(t_grid[0] - t_grid[1])
        nt = len(t_grid)
        r_grid = self.r  # node positions (radial)
        dr_array = np.diff(r_grid)  # spacing between nodes
        nr = len(r_grid)  # number of nodes
        n_el = len(dr_array)  # number of elements

        sigma_n = params_fix['fix_normal_stress'] * (1e6)
        fr_s = params_fix['frac_spacing']
        nu = params_fix['viscosity']

        self.args = (sigma_n, fr_s, dt, nu, dr_array, nr, n_el, Q, r_grid, nt)

        return

    def Cost(self, x, sigma_n, fr_s, dt, nu, dr_array, nr, n_el, Q, r_grid, nt):
        # parameters to be calibrated
        ap_0 = np.power(10., x[0])     # aperture at zero stress
        ap_res = np.power(10., x[1])   # residual aperture
        K_n = np.power(10., x[2])      # fracture normal stiffness
        S = np.power(10., x[3])        # storage
        p_num = self.model(ap_0, ap_res, K_n, S, sigma_n,
                           fr_s, dt, nu, dr_array, nr, n_el, Q, r_grid, nt)
        p_pred = p_num[:, 0] + self.observation[0]
        Coeff_R2 = np.sum(
            np.square(p_pred / 1.e6 - self.observation / 1.e6)) / len(self.observation)

        return Coeff_R2

    def fit(self, TrainingData, TrainingTarget,
            popsize=10,
            polish=True,
            disp=True,
            maxiter=100,
            seed=22,
            tol=1e-5):

        self.translate_training(TrainingData, TrainingTarget, self.params_fix)

        self.result = differential_evolution(self.Cost, self.bounds, args=self.args, popsize=popsize, polish=polish,
                                             disp=disp, maxiter=maxiter, seed=seed, tol=tol)
        self.params = np.power(10., self.result.x[0]), \
            np.power(10., self.result.x[1]), \
            np.power(10., self.result.x[2]), \
            np.power(10., self.result.x[3])

        self.hidden_mechanism = self.predict()

    def predict(self, Plan=None, ):
        if Plan is None:
            Plan = []
        Q = self.Q
        t_grid = Q[:, 0]
        if (len(Plan) > 0):
            Q = np.vstack((Q, Plan))
            # Q = Plan[:,1]
            t_grid = Q[:, 0]
            self.args = list(self.args)
            self.args[-3] = Q[:, 1]
            self.args[3] = abs(t_grid[0] - t_grid[1])
            self.args[-1] = len(t_grid)
        # r_grid = self.r
        pres_num = self.model(self.params[0], self.params[1], self.params[2], self.params[3], self.args[0], self.args[1],
                              self.args[2], self.args[3], self.args[4], self.args[5], self.args[6], self.args[7],
                              self.args[8], self.args[9])
        if len(self.Q) < len(pres_num):
            Prediction = pres_num[len(self.Q):, :] + self.observation[0]
        else:
            Prediction = pres_num + self.observation[0]
        return Prediction


"""
Solving the RADIAL diffusion equation using the implicit finite volume (FV) approach.
"""


@njit(nopython=True)  # activate NUMBA
def flow_model_nonlinear(params, Rw, Np, nu, h, q_f, S, t, dr, x):
    # Here we unpack the parameters
    sigma_n = params[0]         # fixed normal stress
    K_n = params[1]             # fracture normal stiffness
    a_el_max = params[2]        # maximum or zero-stress fracture aperture
    a_res = params[3]           # residual fracture aperture
    fr_s = params[4]            # fracture spacing
    a_shear_f = params[5]       # constant shearing aperture factor
    a_shear_max = params[6]     # maximum shearing aperture
    p_t = params[7]             # threshold pressure
    # scaling factor for shearing aperture magnitude
    alpha = params[8]
    w = params[9]               # weighting parameter for numerical stability
    # setting up empty matrix diagonals and the variables to store
    lwd = np.zeros(Np, dtype=float)
    md = np.zeros(Np + 1, dtype=float)
    upd = np.zeros(Np, dtype=float)
    b = np.zeros(Np + 1, dtype=float)
    pres_step = np.zeros(Np + 1, dtype=float)
    kf_var = np.zeros((len(t), Np + 1), dtype=float)
    a_vals_1 = np.zeros((len(t), Np + 1), dtype=float)
    a_vals_2 = np.zeros((len(t), Np + 1), dtype=float)
    phi_vals = np.zeros((len(t), Np + 1), dtype=float)
    pres_cum = np.zeros((len(t), Np + 1), dtype=float)
    pres_total = np.zeros((len(t), Np + 1), dtype=float)
    # # # set an array full of phi:
    phi = np.full(len(pres_step), 1)
    phi_ = phi  # delete this to include porosity changes
    # # DETECT SHUT-IN #
    # v_inj = np.cumsum(np.maximum(q_f[1:], 0) * np.diff(t))
    # t_shut = t[np.argmax(np.round(v_inj, decimals=5)) + 1]
    # idx_shut = np.where(t_shut==t)[0][0]
    # S_shut = S * 100
    for t_step in range(len(t)):
        # print(t_step)
        # dt = times_step_func(t_step, t)
        dt = abs(t[1] - t[0])
        if dt == 0:
            continue
        # PERMEABILITY CHANGES DUE TO FRACTURE APERTURE VARIATION #
        # Elastic fracture aperture #
        sigma_n_eff = sigma_n - pres_step  # effective normal stress
        # reversible elastic normal closure
        a_el = sigma_n_eff / (K_n + sigma_n_eff / (a_el_max - a_res))
        # mask for all normal effective stress equal to zero or negative
        mask1 = sigma_n_eff <= 0
        a_el[mask1] = 0.0
        # Shearing fracture aperture #
        if t_step == 0:
            pres_cum[t_step] = pres_step
            a_vals_2[t_step] = np.zeros_like(pres_step)
            # PRESSURE DEPENDENT PROSITY CHANGES #
            # phi_ = phi.copy()
            # Cp = S / phi0[0]
        else:
            # set new pressure array
            pres_update = np.copy(pres_cum[t_step - 1])
            # create an array full of zeros
            sf2 = np.zeros_like(pres_step)
            # mask if pressure increases w.r.t. the previous time step
            mask2 = (pres_step - pres_update) / dt > 0
            # modify the array to mimick the step function with mask2
            sf2[mask2] = 1
            # create an array full of zeros
            sf3 = np.zeros_like(pres_step)
            # mask if pressure is above the threshold value
            mask3 = pres_step > p_t
            # modify the array to mimick the step function with mask3
            sf3[mask3] = 1
            a_shear = a_shear_f * sf2 * sf3 * \
                (1 + alpha * (pres_step - pres_update))
            # mask if new (cumulated) shearing aperture is above threshold
            mask4 = (a_vals_2[t_step - 1] + a_shear) > a_shear_max
            # modify the array to mimick the step function with mask4
            a_shear[mask4] = 0
            # add time step contribution to the shear aperture history
            a_vals_2[t_step] = a_vals_2[t_step - 1] + a_shear
            # apply the smoothing factor for numerical stability (here only for a_el)
            a_el = (1 - w) * a_vals_1[t_step - 1] + w * a_el
            # # PRESSURE DEPENDENT POROSITY CHANGES #
            # phi_ = phi * S * (pres_step - pres_update)
            # phi_ = phi + phi_
        # print(f'{phi_[0], t_step, pres_step[0]/1.e6, S}')
        # phi_vals[t_step] = phi_
        # compute the total fracture aperture
        a = a_el_max - a_el + a_vals_2[t_step]
        # save the history of the fracture aperture contributions
        a_vals_1[t_step] = a_el
        # a_vals_2[t_step] = a_shear
        kf = (a**3) / (12 * fr_s)
        # save the permeability history
        kf_var[t_step] = kf
        # boundary conditions at injection point
        md[0] = - kf[0] * 2 * np.pi * Rw * h / \
            (nu * dr[0]) + (S * phi_[0]) / dt
        upd[0] = kf[0] * 2 * np.pi * Rw * h / (nu * dr[0])
        b[0] = - q_f[t_step] + (S * phi_[0]) * pres_step[0] / dt
        for j in range(1, Np):
            r = (x[j - 1] + 0.5 * dr[j - 1]) * dt
            lwd[j - 1] = kf[j - 1] * \
                (r - 0.5 * dr[j - 1]) / (nu * dr[j - 1] * dr[j - 1])
            md[j] = - ((2. * kf[j - 1] * r) / (nu * dr[j - 1]
                       * dr[j - 1]) + (r * (S * phi_[j - 1]) / dt))
            upd[j] = kf[j - 1] * (r + 0.5 * dr[j - 1]) / \
                (nu * dr[j - 1] * dr[j - 1])
            b[j] = - r * (S * phi_[j]) * pres_step[j] / dt
        # boundary conditions at the far field
        lwd[Np - 1] = 0.
        md[Np] = 1.
        upd[Np - 1] = 0.
        b[Np] = 0.
        # Setting up the matrix and solving the system of linear equations
        L = np.zeros((Np + 1, Np + 1))
        for k in range(Np):
            L[k, k] = md[k]
            L[k, k + 1] = upd[k]
            L[k + 1, k] = lwd[k]
        L[Np, Np] = md[Np]
        # if np.linalg.det(L)==0.0:
        #     pdb.set_trace()
        pres_step = np.linalg.solve(L, b)
        # store the new pressure step
        pres_total[t_step] = pres_step
        # print(S)

    return pres_total, kf_var, a_vals_1, a_vals_2, phi_vals


class FlowModel_kchange:
    """
    ADD DESCRIPTION
    """

    def __init__(self,
                 Q,
                 r_grid,
                 h,
                 params_fix,
                 bounds=None,
                 iniguess=(4e-5, 1e-5, 5e-7, 2e-5, 3e-10, 1e-9),
                 forward_model=False):
        if bounds is None:
            bounds = [(1e-5, 1e-2), (1e-8, 1e-5), (1e-8, 1e-4),
                      (1e-6, 1e-4), (1e-12, 1e-7), (-11, -7)]
        self.r = r_grid
        self.h = h
        self.params_fix = params_fix
        self.Q = Q
        self.bounds = bounds
        self.params = iniguess
        self.model = flow_model_nonlinear
        # for callback
        self.best_x = None
        forward_model_ = forward_model,
        self.forward_model = forward_model_[0]
        self.minf = np.inf

    def save_iteration_step(self, x, convergence):
        """
        function to save history of parameters and residual during DEA
        """
        # check if equal
        # np.testing.assert_equal(x, self.best_x)
        self.dea_vals.append([np.power(10., x[0]),
                              np.power(10., x[1]),
                              np.power(10., x[2]),
                              np.power(10., x[3]),
                              np.power(10., x[4]),
                              np.power(10., x[5])])
        self.dea_func.append(self.minf)
        return

    def translate_training(self, TrainingData, TrainingTarget, params_fix):

        self.observation = TrainingTarget[:, 1]
        self.Q = TrainingData[:, :2]

        Q = TrainingData[:, 1]
        t_grid = TrainingData[:, 0]
        self.t_grid = t_grid
        # r_grid = self.r # node positions (radial)
        # dr_array = np.diff(r_grid) # spacing between nodes
        # nr = len(r_grid) - 1 # number of nodes minus one (which is equal to the number of elements)

        self.damping = params_fix['damping_parameter']
        pt = params_fix['threshold_pressure'] * (1e6) - self.observation[0]
        sigma_n = params_fix['fix_normal_stress'] * (1e6)
        fr_s = params_fix['frac_spacing']
        nu = params_fix['viscosity']
        K_n = np.power(10, params_fix['frac_normal_stiffness'])

        self.args = (sigma_n, fr_s, K_n, t_grid, nu, np.diff(
            self.r), len(self.r) - 1, Q, self.r, self.h, self.damping, pt)

        return

    def Cost(self, x, sigma_n, fr_s, K_n, t, nu, dr, nr, Q, r, h, w, pt):
        # parameters to be calibrated
        a_max = np.power(10., x[0])             # aperture at zero stress
        a_res = np.power(10., x[1])             # residual aperture
        # a_shear_init = np.power(10., x[2])      # shear initial aperture
        # gamma = np.power(10., x[3])             # shearing damping factor
        # constant shearing aperture factor
        a_shear_f = np.power(10., x[2])
        # maximum threshold for shearing aperture
        a_shear_max = np.power(10., x[3])
        # pressure difference scaling factor on shearing
        alpha = np.power(10., x[4])
        S = np.power(10., x[5])                 # storage
        # params = (sigma_n, K_n, a_max, a_res, fr_s, a_shear_init, gamma, w)
        params = (sigma_n, K_n, a_max, a_res, fr_s,
                  a_shear_f, a_shear_max, pt, alpha, w)
        p_num = self.model(params, r[0], nr, nu, h, Q, S, t, dr, r)[0]
        p_pred = p_num[:, 0] + self.observation[0]
        Coeff_R2 = np.sum(
            np.square(p_pred / 1.e6 - self.observation / 1.e6)) / len(self.observation)

        if Coeff_R2 < self.minf:
            self.minf = Coeff_R2
            self.best_x = x

        return Coeff_R2

    def fit(self, TrainingData, TrainingTarget,
            popsize=10,
            polish=True,
            disp=True,
            maxiter=20,
            seed=22,
            tol=1e-5):

        self.translate_training(TrainingData, TrainingTarget, self.params_fix)

        # open list to save the progress of the minimization
        self.dea_vals = []
        self.dea_func = []

        self.result = differential_evolution(self.Cost, self.bounds, args=self.args, popsize=popsize, polish=polish,
                                             disp=disp, maxiter=maxiter, seed=seed, tol=tol, updating='immediate', callback=self.save_iteration_step)

        self.params = np.power(10., self.result.x[0]), \
            np.power(10., self.result.x[1]), \
            np.power(10., self.result.x[2]), \
            np.power(10., self.result.x[3]), \
            np.power(10., self.result.x[4]), \
            np.power(10., self.result.x[5])

        self.hidden_mechanism = self.predict()

    def predict(self, Plan=[],):
        Q = self.Q
        t_grid = Q[:, 0]
        if (len(Plan) > 0):
            if self.forward_model == True:
                if len(Q) <= 1:
                    Q = Plan
                else:
                    Q = np.vstack((Q, Plan))
            else:
                Q = np.vstack((Q, Plan))
            t_grid = Q[:, 0]
            self.args = list(self.args)
            self.args[3] = t_grid
            self.args[7] = Q[:, 1]

        # params = (self.args[0], self.args[2], self.params[0], self.params[1], self.args[1], self.params[2], self.params[3], self.args[10])
        # pres_num, self.param_evol_fit_1, self.param_evol_fit_2, self.param_evol_fit_3 = self.model(params, self.args[8][0], self.args[6], self.args[4],
        #                                                                                self.h, self.args[7], self.params[4], self.args[3],
        #                                                                                self.args[5], self.args[8])

        params = (self.args[0], self.args[2], self.params[0], self.params[1], self.args[1],
                  self.params[2], self.params[3], self.args[11], self.params[4], self.args[10])
        pres_num, self.param_evol_fit_1, self.param_evol_fit_2, self.param_evol_fit_3, self.param_evol_fit_4 = self.model(params, self.args[8][0], self.args[6], self.args[4],
                                                                                                                          self.h, self.args[7], self.params[
                                                                                                                              5], self.args[3],
                                                                                                                          self.args[5], self.args[8])

        self.param_evol_pred_1 = self.param_evol_fit_1[len(self.Q):, :]
        self.param_evol_pred_2 = self.param_evol_fit_2[len(self.Q):, :]
        self.param_evol_pred_3 = self.param_evol_fit_3[len(self.Q):, :]
        self.param_evol_pred_4 = self.param_evol_fit_4[len(self.Q):, :]

        if len(self.Q) < len(pres_num):
            if self.forward_model == False:
                # somehow this was a solution before but now it is wrong! (20.08.2024)
                Prediction = pres_num[len(self.Q):, :] + self.observation[0]
            else:
                Prediction = pres_num + self.observation[0]
        else:
            Prediction = pres_num + self.observation[0]
        return Prediction
