"""
Seismicity Models for the HM0/1 classes
by V. Clasen, A. P. Rinaldi, D. Karvounis (2024)
"""

import random

import numpy as np
from numba import njit
from scipy.stats import foldnorm

"""
CAPS model
"""


def Caps_Model_linear(Moments, Edges, MidEdges, max_pressure, CapsSettings, NSim, retriggering=False):
    # get and rename parameters
    sigma1_mean = CapsSettings['inter_sigma1']
    sigma3_mean = CapsSettings['inter_sigma3']
    sigma1_std = CapsSettings['sigma1_std']
    sigma3_std = CapsSettings['sigma3_std']
    cohesion = CapsSettings['cohesion_mean']
    # cohesion_std = CapsSettings['cohesion_std']
    mu = CapsSettings['friction_mean']
    # mu_std = CapsSettings['friction_std']
    # crit = CapsSettings['criticallity']
    stressdrop_coeff = CapsSettings['stressdrop_coeff']
    p_hydro = CapsSettings['phydro']
    seed_density = CapsSettings['seed_density']
    dw = CapsSettings['CAPS_sigma_distr']

    # parameter uncertainties
    if NSim > 1:
        sigma1 = np.random.normal(
            loc=sigma1_mean, scale=0.2 * sigma1_mean * dw, size=NSim)
        sigma3 = np.random.normal(
            loc=sigma3_mean, scale=0.2 * sigma3_mean * dw, size=NSim)
    else:
        sigma1 = np.array([sigma1_mean])
        sigma3 = np.array([sigma3_mean])

    total_events_all = np.zeros((NSim, len(Moments)))
    total_rates_all = np.zeros((NSim, len(Moments) - 1))

    for nrun in range(NSim):
        theta_opt = np.pi / 4. + np.arctan(mu) / 2.

        sigma_n_eff = (sigma1[nrun] + sigma3[nrun]) / 2. + \
            ((sigma1[nrun] - sigma3[nrun]) / 2.) * \
            np.cos(2 * theta_opt) - p_hydro
        tau = np.abs(((sigma1[nrun] - sigma3[nrun]) / 2.)
                     * np.sin(2 * theta_opt))

        # if crit!=0:
        #     # attempt to account for criticallity definition
        #     min_fail_pres = cohesion / (mu - crit) + sigma_n_eff - tau / (mu - crit)
        #     min_fail_pres_std = np.sqrt(
        #        (np.square(sigma1_std) * np.square((1 + np.cos(2 * theta_opt) - np.sin(2 * theta_opt) / (mu - crit)) / 2.))
        #         + np.square(sigma3_std) * np.square((1 - np.cos(2 * theta_opt) + np.sin(2 * theta_opt) / (mu - crit)) / 2.))
        #
        #     crit_check = NormalDist(mu=min_fail_pres, sigma=min_fail_pres_std).cdf(0) <= 0.05
        #
        #     n_crit = crit + crit/10
        #     while crit_check == False:
        #         tau_temp = tau
        #         if (sigma_n_eff - min_fail_pres) > 0:
        #             tau = np.abs(cohesion + (mu - n_crit) * (sigma_n_eff - min_fail_pres))
        #             sigma_n_eff = min_fail_pres + (tau_temp - cohesion) / (mu - n_crit)
        #         else:
        #             tau = np.abs(cohesion + (mu + n_crit) * (sigma_n_eff - min_fail_pres))
        #             sigma_n_eff = min_fail_pres + (tau_temp - cohesion) / (mu + n_crit)
        #         min_fail_pres_temp = min_fail_pres
        #         min_fail_pres = cohesion / (mu - crit) + sigma_n_eff - tau / (mu - crit)
        #         if (min_fail_pres - min_fail_pres_temp) < 1.e-5:
        #             # print('Error: tau cannot get lower than cohesion: criticallity criterion is not met.')
        #             break
        #         crit_check = NormalDist(mu=min_fail_pres, sigma=min_fail_pres_std).cdf(0) <= 0.05
        #         n_crit = n_crit + crit / 10

        fail_pres_mean = cohesion / mu + sigma_n_eff - tau / mu
        # the standard deviation is calculated differently -> check this -> different than for criticallity
        fail_pres_std = np.sqrt(
            (np.square(sigma1_std) * np.square((1
             + np.cos(2 * theta_opt) - np.sin(2 * theta_opt) / mu) / 2.))
            + np.square(sigma3_std) * np.square((1 - np.cos(2 * theta_opt) + np.sin(2 * theta_opt) / mu) / 2.))

        # differential stress:
        # sigma_d = sigma1_eff - sigma3_eff

        # scaling
        seed_number = int(seed_density * np.max(Edges))
        seed_spacebin = seed_number / len(MidEdges)

        fn_cdf = np.zeros_like(max_pressure)
        fn_cdf_rt = np.zeros_like(max_pressure)
        tau_array = np.zeros((len(max_pressure[:, 0]) + 1))
        tau_array[0] = tau
        for tt in range(len(Moments)):
            fn_cdf[tt, :] += foldnorm.cdf(max_pressure[tt, :], np.abs(
                fail_pres_mean) / fail_pres_std, loc=0, scale=fail_pres_std)
            if retriggering == True:
                stressdrop = stressdrop_coeff * \
                    ((tau_array[tt] - cohesion) / mu)
                if np.max(max_pressure[tt, :]) - fail_pres_mean > 0:
                    tau_array[tt + 1] = tau_array[tt] - (
                        stressdrop + mu * (np.max(max_pressure[tt, :]) - fail_pres_mean))
                else:
                    tau_array[tt + 1] = tau_array[tt] - stressdrop
                if tau_array[tt + 1] > tau_array[tt]:
                    tau_array[tt + 1] = tau_array[tt]
                elif tau_array[tt + 1] < 0:
                    tau_array[tt + 1] = 0.
                else:
                    continue
                fail_pres_mean_trig = cohesion / mu + \
                    sigma_n_eff - tau_array[tt + 1] / mu
                n_trig = int(np.sum(fn_cdf[tt, :])) / stressdrop_coeff
                fn_cdf_rt[tt, :] += foldnorm.cdf(max_pressure[tt, :], np.abs(
                    fail_pres_mean_trig) / fail_pres_std, loc=0, scale=fail_pres_std) * n_trig

        fn_cdf = fn_cdf + fn_cdf_rt

        total_events = np.zeros([len(Moments), len(MidEdges)])
        for tt in range(len(Moments)):
            total_events[tt], bins = np.histogram(
                MidEdges, bins=Edges, weights=fn_cdf[tt, :] * seed_spacebin)
        total_rates = np.diff(total_events, axis=0)
        total_events_all[nrun, :] = np.sum(total_events, axis=1)
        total_rates_all[nrun, :] = np.sum(total_rates, axis=1)

    return total_events_all, total_rates_all


"""
Seed Model (after Kiraly et al., 2016)
"""


@njit(nopython=True)
def Seed_Model_linear(Moments, Edges, fm_r, fm_t, pres_t, SeedsSettings, NSim, retriggering=True):

    # np.random.default_rng(seed=np.int64(datetime.now().timestamp()*1.e6))

    # create the seed locations
    max_ext = np.max(Edges)  # furthest distance for the forecast
    seed_density = SeedsSettings[0]  # seed density
    seed_number = int(seed_density * max_ext)  # total number of seeds
    Rcoord = max_ext * np.random.rand(seed_number * NSim)  # placement of seeds

    # get and rename parameters
    sigma1_mean = SeedsSettings[1]
    sigma3_mean = SeedsSettings[2]
    sigma1_std = SeedsSettings[3]
    sigma3_std = SeedsSettings[4]
    cohesion = SeedsSettings[5]
    # cohesion_std = SeedsSettings[5]
    mu = SeedsSettings[6]
    # mu_std = SeedsSettings[2]
    crit = SeedsSettings[7]
    # theta_std = SeedsSettings[10]
    stressdrop_coeff = SeedsSettings[8]
    p_hydro = SeedsSettings[9]

    # b-values and corresponding differential stresses
    b_min = SeedsSettings[10]
    b_max = SeedsSettings[11]
    b_min_diff_stress = SeedsSettings[12]
    b_max_diff_stress = SeedsSettings[13]

    Mc = SeedsSettings[14]

    # setting arrays to be calculated
    # optimal orientation of each seed
    theta_opt = np.pi / 4. + np.arctan(mu) / 2.
    sigma1 = sigma1_mean + sigma1_std * np.random.randn(seed_number * NSim)
    sigma3 = sigma3_mean + sigma3_std * np.random.randn(seed_number * NSim)

    sig_check = np.where((sigma3 - p_hydro) > (sigma1 - p_hydro))
    while len(sig_check[0]):
        sigma1[sig_check] = sigma1_mean + sigma1_std * \
            np.random.randn(len(sig_check[0]))
        sigma3[sig_check] = sigma3_mean + sigma3_std * \
            np.random.randn(len(sig_check[0]))
        sig_check = np.where((sigma3 - p_hydro) > (sigma1 - p_hydro))

    sigma_n_eff = ((sigma1 + sigma3) / 2. + ((sigma1 - sigma3) / 2.)
                   * np.cos(2. * theta_opt)) - p_hydro
    tau = np.abs((sigma1 - sigma3) / 2. * np.sin(2. * theta_opt))

    crit_check = np.where(cohesion / (mu - crit)
                          + sigma_n_eff - tau / (mu - crit) < 0.)
    while len(crit_check[0]):
        tau_temp = tau[crit_check]
        tau[crit_check] = np.abs(
            cohesion + (mu - crit) * sigma_n_eff[crit_check])
        sigma_n_eff[crit_check] = (tau_temp - cohesion) / (mu - crit)
        crit_check = np.where(cohesion / (mu - crit)
                              + sigma_n_eff - tau / (mu - crit) < 0.)

    fail_pressure = cohesion / mu + sigma_n_eff - tau / mu

    # MidEdges = np.zeros(len(Edges) - 1)
    # for i in range(len(MidEdges)):
    #     MidEdges[i] = (Edges[i] + Edges[i + 1]) / 2.

    diff_stress = (sigma1 - sigma3) - p_hydro

    # open lists of parameters:
    b_val = []
    mag = []
    time = []
    R = []
    pressure = []
    trig_index = []
    differential_stress = []
    sig1 = []
    sig3 = []
    stress_drop = []

    Rind = np.argmin(
        np.abs(fm_r - np.reshape(Rcoord, (len(Rcoord), 1))).T, axis=0)

    # comment previous and uncomment next 3 lines to make it work in a different environment than in the pycharm setup
    # loc_dist = np.abs(fm_r - np.reshape(Rcoord, (len(Rcoord), 1)))
    # Rind = np.empty((len(Rcoord)))
    # for i in range(len(Rcoord)): Rind[i] = (np.argmin(loc_dist[i]))

    events_map = np.ravel((np.ones((seed_number, NSim))
                          * np.array(list(range(0, NSim)))))

    pres_map = np.empty((len(Moments), len(Rind)))
    for j in range(len(Rind)):
        pres_map[:, j] = pres_t[int(Rind[j]), :]

    # total_events = np.zeros((NSim, len(Moments), len(MidEdges)))
    total_events = np.zeros((NSim, len(Moments)))

    t_events = 0
    tt = 0

    for t in range(len(Moments)):
        # for t in Moments:
        OverShootPressure = pres_map[t] - fail_pressure
        # OverShootPressure = pres_func(Rcoord, t) - fail_pressure
        trig_event_tbin = np.where(OverShootPressure > 0)
        trig_event_total_array = trig_event_tbin[0]

        if len(trig_event_total_array):
            # go over all triggered seeds
            for i in np.nditer(trig_event_total_array):
                # if differential stress is higher than the differential stress corresponding to b_min
                if diff_stress[i] > b_min_diff_stress:
                    b_value = b_min
                else:
                    # otherwise find the corresponding b-value
                    mm = ((b_max - b_min)
                          / (b_max_diff_stress - b_min_diff_stress))
                    qq = b_max - mm * b_max_diff_stress
                    b_value = mm * diff_stress[i] + qq

                # idx_pos = np.argmin(np.abs(MidEdges - Rcoord[i]))
                # total_events[int(event_map[i]), t_events, idx_pos] += 1
                total_events[int(events_map[i]), t_events] += 1

                # save for each triggered seed into the catalog the...
                # b-value
                b_val.append(b_value)
                # magnitude
                mag.append(
                    Mc + np.random.exponential(1. / (b_value * np.log(10))))
                # time
                time.append(t)
                # x-location
                R.append(Rcoord[i])
                # pressure
                pressure.append(OverShootPressure[i] + fail_pressure[i])
                # triggering index
                trig_index.append(float(i + 1))
                # differential stress
                differential_stress.append(diff_stress[i])
                # maximum effective principal stress
                sig1.append(sigma1[i] - p_hydro)
                # minimum effective principal stress
                sig3.append(sigma3[i] - p_hydro)

                if retriggering == True:
                    "1)"
                    # the stress drop is proportional to the normal stress
                    stressdrop = stressdrop_coeff * \
                        (sigma_n_eff[i] - fail_pressure[i])
                    # stressdrop = stressdrop_coeff * (((tau[i] - cohesion) / mu))
                    # shear stress on seeds after triggering
                    tau[i] = tau[i] - (stressdrop + mu * OverShootPressure[i])
                    # change/reduction in the maximum principal stress
                    sigma1[i] = sigma1[i] - 2 * \
                        (stressdrop + mu
                         * OverShootPressure[i]) / np.abs(np.sin(2 * theta_opt))
                    # differential stress on seeds after triggering
                    diff_stress[i] = (sigma1[i] - sigma3[i]) - p_hydro
                    # normal effective stress on seeds after triggering (no effect for now)
                    # sigma_n_eff[i] = (sigma1[i] + sigma3[i])/2. + ((sigma1[i] - sigma3[i]) * np.cos(2*theta_opt) / 2.) - p_hydro
                    fail_pressure[i] = cohesion / mu + \
                        sigma_n_eff[i] - tau[i] / mu
                    # stressdrop
                    stress_drop.append(stressdrop)
                else:
                    fail_pressure[i] = random.random() * \
                        np.square(np.max(pres_t))
                    # The following lines are commented (not really necessary) as they do not work with NUMBA...
                    # ind = np.where(fail_pressure[i] < np.max(pres_sol))
                    # while len(ind[0]):
                    #     fail_pressure[i] = random.random() * np.square(np.max(pres_sol))
                    #     ind = np.where(fail_pressure[i] < np.max(pres_sol))
                    stress_drop.append(0)

        t_events = t_events + 1
        if not t_events >= len(Moments):
            # total_events[:, t_events, :] = total_events[:, t_events - 1, :]
            total_events[:, t_events] = total_events[:, t_events - 1]
        tt = tt + 1

    EQ = (b_val, mag, time, R, pressure, trig_index,
          differential_stress, sig1, sig3, stress_drop)
    total_rates = np.diff(total_events)
    # total_events_noloc = np.sum(total_events, axis=2)
    # total_rates_noloc = np.diff(total_events_noloc, axis=1)

    return EQ, total_events, total_rates


"""
COMMENTS:
1) The version: stressdrop = stressdrop_coeff * (((tau[i] - cohesion) / mu) + fail_pressure[i]),
is not correct because the stressdrop is related to the normal stress (not effective normal stress and also not to the
normal stress reduction necessary to reach failure). Then, the stress drop is directly related to the normal stress but
the magnitude of reduction in the shear stress is actually related also to the amount of overpressure for that given seed.
If the overpressure is large, tau will experience a larger reduction.
2)
"""
