import copy

import numba as nb
import numpy as np
from FlowModel import (FlowModel0, FlowModel1, FlowModel2, FlowModel3,
                       FlowModel4)
from scipy.interpolate import CloughTocher2DInterpolator
from scipy.optimize import differential_evolution
from SeismicityModel import Caps_Model_linear as CAPS
from SeismicityModel import Seed_Model_linear as SEED
from utils import (Calibrate_Observed_GR, LL_cost_function_1D,
                   LS_cost_function_1D, Refine_An_Edge,
                   b_value_Lasso_calibration_1D)


class HM01:
    def __init__(self,
                 TrainingData,
                 time_col='time',
                 flow_col='flow',
                 press_col='pressure',
                 source_times_col='ts',
                 hypocenter_cols=['Rs'],
                 Mw_cols='mag',
                 dM=0.01,
                 max_iteration=-1,
                 cost_fun='LS',
                 inv_mod='seq',
                 seis_model='CAPS',
                 retriggering=False,
                 forward_model=False,  # True: FWD, False: Inversion mode
                 flow_model='line_source',
                 params_method='Chi2crt',
                 NSim=1,
                 seed_model_version='1',
                 workers=1,
                 opt_algorithm='DE',
                 max_grid_ext=300,
                 layer_thick=100,
                 cross_area=94,
                 sensitivity_analysis=False,
                 init_params=[],
                 Settings=[]):

        self.Columns = [time_col, flow_col, press_col]

        self.opt_algorithm = opt_algorithm
        self.sensitivity_analysis = sensitivity_analysis
        self.workers = workers
        self.retriggering = retriggering
        self.forward_model = forward_model

        # self.TrainingData = None
        self.TrainingData = TrainingData
        self.training_catalog = None

        self.ts = source_times_col
        self.Xs = hypocenter_cols
        self.Mw = Mw_cols
        self.mc_bin_size = dM

        self.cost_fun = cost_fun
        listOfStrings = ['LL', 'LS']
        if self.cost_fun not in listOfStrings:
            print("Cost function '", self.cost_fun, "' not found. Possible values are:",
                  listOfStrings, ". Using default mode (LL).")
            self.cost_fun = 'LL'

        self.inv_mod = inv_mod
        listOfStrings = ['seq', 'joint']
        if self.inv_mod not in listOfStrings:
            print("Inversion mode '", self.cost_fun, "' not found. Possible values are:",
                  listOfStrings, ". Using default mode (seq).")
            self.inv_mod = 'seq'

        self.flw_model = flow_model
        listOfStrings = ['point_source', 'line_source',
                         'num_const_perm', 'num_var_perm', 'num_var_perm_FV']
        if self.flw_model not in listOfStrings:
            print("Fluid model '", self.cost_fun, "' not found. Possible values are:",
                  listOfStrings, ". Using default mode (line_source).")
            self.flw_model = 'line_source'

        self.NSim = NSim
        self.seed_model_version = seed_model_version
        self.workers = workers
        self.seis_model = seis_model
        self.params_method = params_method

        self.seed_settings = Settings[0]
        self.seis_bounds = Settings[1]
        self.flow_bounds = Settings[2]
        self.ExSolSettings = Settings[3]
        self.CAPSSettings = Settings[4]
        self.SEEDSettings = Settings[4]

        self.CostFunction = None
        self.LL_Edges = None
        self.LL_Moments = None
        self.Training_MaxPressure = None
        self.Training_XYZ = None
        self.Training_Volume = None

        # input for the flow models
        self.max_grid_ext = max_grid_ext
        self.layer_thick = layer_thick
        self.cross_area = cross_area

        # upper b-value limit
        self.maxbval_bounds = [(1., self.seed_settings["b_max"])]

        if self.retriggering == True:
            if self.seed_settings['criticallity'] != 0:
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'stressdrop_coeff',
                                     'cohesion_mean', 'friction_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['stressdrop_coeff'][0],
                                    self.seis_bounds['stressdrop_coeff'][1]),
                                   # (self.seis_bounds['criticallity'][0], self.seis_bounds['criticallity'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False, False, False, False]
            else:
                # self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'stressdrop_coeff', 'friction_mean', 'seed_density']
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'stressdrop_coeff',
                                     'cohesion_mean', 'friction_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['stressdrop_coeff'][0],
                                    self.seis_bounds['stressdrop_coeff'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False, False, False, False]
        else:
            if self.seed_settings['criticallity'] != 0:
                # self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'friction_mean', 'seed_density']
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std',
                                     'friction_mean', 'cohesion_mean', 'seed_density']
                # self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'criticallity', 'friction_mean', 'cohesion_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   # (self.seis_bounds['criticallity'][0], self.seis_bounds['criticallity'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                # self.logscale = [False, False, False, False]
                self.logscale = [False, False, False, False, False]
                # self.logscale = [False, False, False, False, False, False]
            else:
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std',
                                     'cohesion_mean', 'friction_mean', 'seed_density']
                # self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'friction_mean', 'seed_density']
                # self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'cohesion_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False, False, False]
                # self.logscale = [False, False, False, False]

        """
        4 different flow models can be choosen:
        -> point_source: analytical point source solution. Two parameters, storage coefficient and permeability, are calibrated.
        -> line_source: Theis' solution. Two parameters, storage coefficient and permeability, are calibrated.
        -> num_const_perm: numerical solution without permeability changes. Only the permeability is calibrated and is a fixed value.
        -> num_var_perm: numerical solution with permeability changes. Three parameters are calibrated that control the changes in permeability with pressure.
        """

        if self.flw_model == 'point_source':
            self.OptimizeKeys = self.OptimizeKeys + \
                ['storage', 'permeability']  # parameters to calibrate
            self.FlowBounds = [(self.flow_bounds['storage'][0], self.flow_bounds['storage'][1]),
                               (self.flow_bounds['permeability'][0], self.flow_bounds['permeability'][1])]
            self.logscale = self.logscale + [True, True]
            # activate the flow model:
            self.flow_model = FlowModel0(self.TrainingData[self.Columns[:2]].values,
                                         np.geomspace(
                                             0.1, self.max_grid_ext, 30, endpoint=True), self.cross_area,
                                         bounds=self.FlowBounds)

        elif self.flw_model == 'line_source':
            self.OptimizeKeys = self.OptimizeKeys + ['storage', 'permeability']
            self.FlowBounds = [(self.flow_bounds['storage'][0], self.flow_bounds['storage'][1]),
                               (self.flow_bounds['permeability'][0], self.flow_bounds['permeability'][1])]
            self.logscale = self.logscale + [True, True]
            # activate the flow model:
            self.flow_model = FlowModel1(self.TrainingData[self.Columns[:2]].values,
                                         np.geomspace(
                                             0.1, self.max_grid_ext, 30, endpoint=True), self.layer_thick,
                                         bounds=self.FlowBounds, iniguess=(self.seed_settings['storage'], self.seed_settings['permeability']))

        elif self.flw_model == 'num_const_perm':
            self.OptimizeKeys = self.OptimizeKeys + ['storage', 'permeability']
            self.FlowBounds = [(self.flow_bounds['storage'][0], self.flow_bounds['storage'][1]),
                               (self.flow_bounds['permeability'][0], self.flow_bounds['permeability'][1])]
            self.logscale = self.logscale + [True, True]
            # activate the flow model:
            self.flow_model = FlowModel2(self.TrainingData[self.Columns[:2]].values,
                                         np.geomspace(
                                             0.1, self.max_grid_ext, 30, endpoint=True), self.layer_thick,
                                         self.seed_settings, bounds=self.FlowBounds)

        elif self.flw_model == 'num_var_perm':  # HM1 class with FD
            self.OptimizeKeys = self.OptimizeKeys + \
                ['zero_stress_aperture', 'residual_aperture',
                    'frac_normal_stiffness', 'storage']  # parameters to calibrate
            # self.FlowBounds = [(self.seed_settings['zero_stress_aperture'], -2), # zero-stress aperture
            #                    (self.seed_settings['residual_aperture'], -7),   # residual aperture
            #                    (self.seed_settings['frac_normal_stiffness'], 14),    # fracture normal stiffness
            #                    (-9, -7)]  # storage
            self.FlowBounds = [(self.flow_bounds['zero_stress_aperture'][0], self.flow_bounds['zero_stress_aperture'][1]),
                               (self.flow_bounds['residual_aperture'][0],
                                self.flow_bounds['residual_aperture'][1]),
                               (self.flow_bounds['frac_normal_stiffness'][0],
                                self.flow_bounds['frac_normal_stiffness'][1]),
                               (self.flow_bounds['storage'][0], self.flow_bounds['storage'][1])]
            self.logscale = self.logscale + [True, True, True, True]
            # activate the flow model:
            self.flow_model = FlowModel3(self.TrainingData[self.Columns[:2]].values,
                                         np.geomspace(
                                             0.1, self.max_grid_ext, 30, endpoint=True), self.layer_thick,
                                         self.seed_settings, bounds=self.FlowBounds)

        else:  # HM1 cass with FV
            # self.OptimizeKeys = self.OptimizeKeys + ['zero_stress_aperture', 'residual_aperture', 'shearing_aperture_factor', 'shear_scaling_factor', 'storage'] # parameters to calibrate
            self.OptimizeKeys = self.OptimizeKeys + ['zero_stress_aperture', 'residual_aperture',
                                                     'shearing_aperture_factor', 'max_shear_frac_aperture', 'shear_scaling_factor', 'storage']
            self.FlowBounds = [(self.flow_bounds['zero_stress_aperture'][0], self.flow_bounds['zero_stress_aperture'][1]),
                               (self.flow_bounds['residual_aperture'][0],
                                self.flow_bounds['residual_aperture'][1]),
                               (self.flow_bounds['shearing_aperture_factor'][0],
                                self.flow_bounds['shearing_aperture_factor'][1]),
                               (self.flow_bounds['max_shear_frac_aperture'][0],
                                self.flow_bounds['max_shear_frac_aperture'][1]),
                               (self.flow_bounds['shear_scaling_factor'][0],
                                self.flow_bounds['shear_scaling_factor'][1]),
                               (self.flow_bounds['storage'][0], self.flow_bounds['storage'][1])]
            # self.logscale = self.logscale + [True, True, True, True, True]
            self.logscale = self.logscale + \
                [True, True, True, True, True, True]
            # activate the flow model:
            self.flow_model = FlowModel4(self.TrainingData[self.Columns[:2]].values,
                                         np.geomspace(
                                             0.1, self.max_grid_ext, 30, endpoint=True), self.layer_thick,
                                         self.seed_settings, bounds=self.FlowBounds, forward_model=forward_model,
                                         iniguess=(self.seed_settings['zero_stress_aperture'], self.seed_settings['residual_aperture'],
                                                   self.seed_settings['shearing_aperture_factor'], self.seed_settings[
                                                       'max_shear_frac_aperture'],
                                                   self.seed_settings['shear_scaling_factor'], self.seed_settings['storage']))

        if seis_model == 'SEED':
            self.SEEDmodelKeys = ['seed_density', 'inter_sigma1', 'inter_sigma3', 'sigma1_std', 'sigma3_std',
                                  'cohesion_mean', 'friction_mean', 'criticallity', 'stressdrop_coeff', 'phydro',
                                  'b_min', 'b_max', 'b_min_diff_stress', 'b_max_diff_stress', 'Mag_Completeness']

        self.disp = True
        self.maxiter = max_iteration
        self.seed = 42

        # for callback
        self.best_x = None
        self.minf = np.inf

        # self.init_params = [5, 0.2, 5.2, 0.44]

        # currently not used:
        # if self.opt_algorithm=='LMFIT':
        #     params = Parameters()
        #     for i in range(len(self.OptimizeKeys)-1):
        #         params.add(self.OptimizeKeys[i], self.init_params[i], min=self.SEISbounds[i][0], max=self.SEISbounds[i][1])
        #     self.SEISparams = params

    def update_seeds_settings(self):
        # would you rather the Mc was calculated or gained
        # from the QuakeMl data?
        Mc, b_value = Calibrate_Observed_GR(
            self.training_catalog, self.mc_bin_size, self.Mw)
        self.seed_settings["Mag_Completeness"] = Mc
        self.seed_settings["b_value"] = b_value
        self.training_catalog = self.training_catalog[
            self.training_catalog[self.Mw] >= Mc]

        if (self.seed_settings["b_vs_depth_range"]):
            a, m = b_value_Lasso_calibration_1D(
                np.array(self.training_catalog[self.Xs].values),
                np.array(self.training_catalog[[self.Mw]])[:, 0],
                bins=100)
        else:
            a, m = b_value, np.zeros(1)

        self.seed_settings["a"] = a
        self.seed_settings["m"] = m

    def Seed_max_bval_fit(self, x):
        b_min = self.seed_settings['b_min']
        # to be optimized
        b_max = x
        self.seed_settings['b_max'] = b_max
        b_max_diff_stress = self.seed_settings['b_max_diff_stress']
        b_min_diff_stress = self.seed_settings['b_min_diff_stress']
        # mean differential stress from seeds
        diff_stress_mean = self.seed_settings['diff_stress_mean']
        # b-value from observed data
        b_val_obs = self.seed_settings['b_value']

        mm = ((b_max - b_min) / (b_max_diff_stress - b_min_diff_stress))
        qq = b_max - mm * b_max_diff_stress
        b_val_model = mm * diff_stress_mean + qq

        b_val_diff = np.abs(b_val_obs - b_val_model)

        return b_val_diff

    def ll_score(self, x):
        y = self.CostFunction.y
        log_factorial_y = self.CostFunction.logyfact
        TestSettings = copy.deepcopy(self.seed_settings)

        for i in range(len(self.SEISbounds)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])

        y0 = CAPS.Caps_Model_linear(self.LL_Moments[1:], self.LL_Edges[self.LL_Edges >= 0.], self.Training_R,
                                    self.Training_MaxPressure, TestSettings, NSim=1, retriggering=self.retriggering)[0]

        LL = np.sum(y * np.log(y0) - y0 - log_factorial_y) / \
            len(self.training_catalog)  # Poissonian LL
        # LL = np.sum(y * np.log10(y1) - y1 - log_factorial_y) / float(len(y1))
        if np.isfinite(LL):
            return -LL
        return 1.e20

    def ls_score(self, x):
        y = self.CostFunction.y
        TestSettings = copy.deepcopy(self.seed_settings)

        for i in range(len(self.SEISbounds)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])

        y0 = CAPS.Caps_Model_linear(self.LL_Moments[1:], self.LL_Edges[self.LL_Edges >= 0.], self.Training_R,
                                    self.Training_MaxPressure, TestSettings, NSim=1, retriggering=self.retriggering)[0]

        LS = np.sum(np.square(y - y0)) / len(y)
        if np.isfinite(LS):
            if LS < self.minf:
                self.minf = LS
                self.best_x = x
            return LS
        return 1.e20

    def ls_score_seed(self, x):
        y = self.CostFunction.y
        TestSettings = copy.deepcopy(self.seed_settings)

        for i in range(len(self.SEISbounds)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])

        # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
        SeedsSettings = nb.typed.List()
        for j in range(len(self.SEEDmodelKeys)):
            SeedsSettings.append(float(TestSettings[self.SEEDmodelKeys[j]]))

        y0 = SEED.Seed_Model_linear(self.LL_Moments[1:], self.LL_Edges[self.LL_Edges >= 0.],
                                    self.flow_model.r, self.flow_model.Q[:, 0],
                                    self.pres_t_fit,
                                    SeedsSettings, self.Nreal, self.retriggering)[1]

        y1 = np.median(y0, axis=0)

        LS = np.sum(np.square(y - y1)) / len(y)

        if np.isfinite(LS):
            if LS < self.minf:
                self.minf = LS
                self.best_x = x
            return LS
        return 1.e20

    def ll_score_joint(self, x):

        TestSettings = copy.deepcopy(self.seed_settings)
        # print(x)
        for i in range(len(x)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])
        # end_time = time.time()

        # self.flow_model.params = TestSettings[self.OptimizeKeys[4]], TestSettings[self.OptimizeKeys[5]]
        self.flow_model.params = TestSettings[self.OptimizeKeys[4]]

        self.flow_model.hidden_mechanism = self.flow_model.predict()

        Coeff_R2 = np.sum(np.square(self.flow_model.hidden_mechanism[:, 0] / 1.e6
                                    - self.flow_model.observation.T / 1.e6)) / len(self.flow_model.observation)

        self.timesteps = self.TrainingData.time.values
        inipres = self.TrainingData[self.Columns[-1]].values[0]
        self.PressureFunction = self.create_interpolator(
            self.timesteps, (self.flow_model.hidden_mechanism - inipres) / 1.e6)

        (self.Training_MaxPressure,
         self.Training_XYZ,
         self.Training_Volume) = self.Interpolate_1D(self.LL_Moments, self.LL_Edges,
                                                     self.ExSolSettings, self.CAPSSettings)

        y1 = CAPS.Forecast_BinnedRate_only_1D(
            self.LL_Moments[1:], self.LL_Edges, self.Training_MaxPressure,
            self.Training_XYZ, self.Training_Volume, TestSettings)
        # end_time = time.time() - end_time
        # print("Elapsed time to run Forecast_Binned_rate was %g seconds" % (end_time))

        y0 = np.diff(y1, axis=(0))
        y1 = np.zeros(y1.shape)
        y1[1:, :] = y0
        y1 = y1.ravel()
        y1[np.where(y1 <= 0.)] = 1.e-20

        y = self.SeisCostFunction.y
        log_factorial_y = self.SeisCostFunction.logyfact

        LL = -np.sum(y * np.log(y1) - y1 - log_factorial_y) / \
            len(self.training_catalog)
        Cost = LL + Coeff_R2
        # LL = np.sum(y * np.log10(y1) - y1 - log_factorial_y) / float(len(y1))
        if np.isfinite(Cost):
            return Cost
        return 1.e20

    def ls_score_joint(self, x):

        TestSettings = copy.deepcopy(self.seed_settings)
        # print(x)
        for i in range(len(x)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])
        # end_time = time.time()

        self.flow_model.params = TestSettings[self.OptimizeKeys[4]
                                              ], TestSettings[self.OptimizeKeys[5]]
        # self.flow_model.params = TestSettings[self.OptimizeKeys[4]]

        self.flow_model.observation = self.TrainingData[self.Columns[::2]].values[:, 1:]

        self.flow_model.hidden_mechanism = self.flow_model.predict()

        Coeff_R2 = np.sum(np.square(self.flow_model.hidden_mechanism[:, 0] / 1.e6
                                    - self.flow_model.observation.T / 1.e6)) / np.max(self.flow_model.observation / 1.e6)

        self.timesteps = self.TrainingData.time.values
        inipres = self.TrainingData[self.Columns[-1]].values[0]
        self.PressureFunction = self.create_interpolator(
            self.timesteps, (self.flow_model.hidden_mechanism - inipres) / 1.e6)

        (self.Training_MaxPressure,
         self.Training_XYZ,
         self.Training_Volume) = self.Interpolate_1D(self.LL_Moments, self.LL_Edges,
                                                     self.ExSolSettings, self.CAPSSettings)

        y0 = CAPS.Forecast_BinnedRate_only_1D(
            self.LL_Moments[1:], self.LL_Edges, self.Training_MaxPressure,
            self.Training_XYZ, self.Training_Volume, TestSettings)

        y1 = np.sum(y0, axis=1)

        # end_time = time.time() - end_time
        # print("Elapsed time to run Forecast_Binned_rate was %g seconds" % (end_time))

        y = self.SeisCostFunction.y

        LS = np.sum(np.square(y - y1)) / len(y)

        Cost = LS + Coeff_R2
        # LL = np.sum(y * np.log10(y1) - y1 - log_factorial_y) / float(len(y1))
        if np.isfinite(Cost):
            return Cost
        return 1.e20

    def save_iteration_step(self, x, convergence):
        """
        function to save history of parameters and residual during DEA
        """
        # check if equal
        # np.testing.assert_equal(x, self.best_x)
        self.dea_vals.append(x)
        self.dea_func.append(self.minf)
        return

    def create_interpolator(self, t_grid, solution):
        if len(self.flow_model.r) == len(solution[0]):
            mesh = np.meshgrid(self.flow_model.r, t_grid)
        else:
            mesh = np.meshgrid(self.flow_model.r[1:], t_grid)
        interpd = CloughTocher2DInterpolator(
            np.array([mesh[0].ravel(), mesh[1].ravel()]).T, solution.ravel(), fill_value=0)
        return interpd

    def mask_for_median(self, events_grid):

        total_events_spacesum = np.sum(events_grid, axis=0)
        total_events_median_new = np.zeros_like(total_events_spacesum)
        total_events_median_new_single = np.zeros_like(
            total_events_spacesum[:, 0])

        for i in range(len(total_events_spacesum)):
            if np.sum(total_events_spacesum[i]) == 0:
                total_events_median_new[i] = total_events_spacesum[i]
                total_events_median_new_single[i] = np.median(
                    total_events_median_new[i], axis=0)
            else:
                masked = np.ma.masked_where(
                    total_events_spacesum[i] == 0, total_events_spacesum[i])
                total_events_median_new_single[i] = np.ma.median(masked)

        return total_events_median_new_single

    def Interpolate_1D(self, Forecasts_Moments, Forecasts_Edges, CAPS_KD_Settings):
        """
        Same as CAPS.InterpolateMaximumPressure, but using a self.PressureFunction created by the fit
        """
        RR = CAPS_KD_Settings["RefinementRatio"]
        # for 1D we use directly the cartesian grid as the radius
        MyR = Refine_An_Edge(Forecasts_Edges, RR)
        LaterKeepIndices = np.logical_and(np.isfinite(MyR), MyR > 0.)
        My_Working_R = MyR[LaterKeepIndices]
        DT = np.concatenate((self.timesteps, Forecasts_Moments))
        MyT = np.sort(np.unique(DT))
        MyPressure, MaxPressureAll = np.zeros([2, len(MyT), len(My_Working_R)])
        PM = np.zeros(len(My_Working_R))
        for timeIter in range(len(MyT)):
            ConsiderTime = MyT[timeIter]
            MyPressure[timeIter, :] = self.PressureFunction(
                My_Working_R, ConsiderTime)
            MaxPressureAll[timeIter, :] = np.maximum(
                PM, MyPressure[timeIter, :])
            PM = MaxPressureAll[timeIter, :]

        TimeBinIndeces = np.intersect1d(Forecasts_Moments, MyT, return_indices=True)[
            2]  # discard unecessary time discretization now
        MaxPressure = MaxPressureAll[TimeBinIndeces]
        dV = np.unique(np.diff(Forecasts_Edges, axis=0))
        # dV=np.unique(np.diff(Forecasts_Edges, axis=0), axis=0)
        # if(len(dV)>1):
        # print ("Warning: non equidistant edges have been passed in the one direction. Correct array of control distances.")
        # \todo: assume equidistant edges and refinement
        My_Cell_Width = np.prod(np.amin(dV, axis=0) / RR)
        return MaxPressure, My_Working_R, np.ones(len(My_Working_R)) * My_Cell_Width

    def ls_params_score(self, percent):

        par_grd_2 = []
        for p in range(len(self.CAPSbounds)):
            # par_distr = self.Params_CAPS(forecast_edges, max_pressure_past, working_xyz_past, cells_volume_past, p)
            par_distr = self.Params_CAPS_2(percent, p)
            par_grd_2.append(par_distr)
        cal_val_2 = np.median(par_grd_2, axis=1)

        LS = np.sum(np.square(self.cal_val - cal_val_2)) / len(self.cal_val)

        if np.isfinite(LS):
            return LS
        return 1.e20

    # def sensitivity_analysis_fun(self, params_distr, z_i, f_z_i, forecast_edges, method='sobol_global'):
    #
    #     if method == 'sobol_global':
    #         """
    #             GLOBAL SENSITIVITY ANALYSIS
    #
    #             Ref. -> E. Qian, B. Peherstorfer, D. O'Malley, V. V. Vesselinov, and K. Willcox.
    #                     Multifidelity Monte Carlo Estimation of Variance and Sensitivity Indices
    #                     SIAM/ASA J. Uncertainty Quantification. Vol. 6, No. 2, pp. 683-706
    #
    #             Explanation parameters:
    #             params_distr -> distribution of the input parameters
    #             z_i -> values of parameters from FIRST random sampling (from NSim Monte Carlo simulations)
    #             f_z_i -> output of the model from z_i
    #             z_i_ -> values of input parameters from SECOND random sampling (from NSim Monte Carlo simulations)
    #             f_z_i_ -> output of the model from z_i_
    #             y_i_j -> same as z_i_ but fixing the one input parameter of interest from z_i
    #             f_y_i_j -> output of the model from y_i_j
    #
    #             E -> sample mean (mean of the entire output, over all times and space)
    #             V -> variance of all inputs
    #             V_j -> variance contributed by the input j
    #             T_j -> total variance contributed by the input j
    #             s_j -> Sobol' main sensitivity index for input j (portion of the output that can be attributed to the influence of the input j only, fraction of variance of input j alone)
    #                     also: first order sensitivity indices
    #             s_j_t -> Sobol' total effect sensitivity index for input j (additionally takes into account the sum of the contributions)
    #                     also: total-order sensitivity indices
    #         """
    #
    #         z_i = np.array(z_i).T
    #
    #         f_z_i_copy = np.array(f_z_i)
    #         f_z_i_new = []
    #         for l in range(len(f_z_i)):
    #             f_z_i_new.append(np.sum(f_z_i_copy[l][-1]))
    #         f_z_i = np.array(f_z_i_new)
    #         del f_z_i_copy, f_z_i_new
    #         # f_z_i = np.sum(np.sum(np.array(f_z_i), axis=2), axis=1)
    #
    #         E = np.sum(f_z_i) / self.NSim
    #
    #         V = (np.sum((f_z_i - E) ** 2)) / (self.NSim - 1)
    #
    #         pool = mp.Pool(self.nr_cpu)
    #         result = pool.starmap(self.Parallel_CAPS_fitting,
    #                               [(nr_sim, self.TrainingData.time.values[1:], forecast_edges, self.par_grd) for nr_sim
    #                                in range(self.NSim)])
    #         total_events = [result[a][0] for a in range(len(result))]
    #         params = [result[a][1] for a in range(len(result))]
    #
    #         z_i_ = np.array(params).T
    #
    #         total_events_copy = np.array(total_events)
    #         total_events_new = []
    #         for l in range(len(total_events)):
    #             total_events_new.append(np.sum(total_events_copy[l][-1]))
    #         f_z_i_ = np.array(total_events_new)
    #         # f_z_i_ = np.sum(np.sum(np.array(total_events), axis=2), axis=1)
    #
    #         s_j_array = []
    #         s_j_t_array = []
    #
    #         for p in range(len(self.CAPSbounds)):
    #             y_i_j = np.copy(z_i_)
    #             # y_i_j = np.copy(z_i)
    #             y_i_j[p, :] = z_i[p, :]
    #             # y_i_j[p,:] = np.full_like(z_i[0], self.seed_settings_optimal[self.OptimizeKeys[p]])
    #             # y_i_j[p, :] = z_i_[p, :]
    #             random_method = False
    #             total_events_array = []
    #             for i in range(len(y_i_j[0])):
    #                 for j in range(len(self.CAPSbounds)):
    #                     self.seed_settings[self.OptimizeKeys[j]] = y_i_j[j, i]
    #                 total_events = CAPS.Forecast_BinnedRate_only_1D(self.TrainingData.time.values[1:],
    #                                                                 forecast_edges[0], self.max_pressure_past,
    #                                                                 self.working_xyz_past, self.cells_volume_past,
    #                                                                 self.seed_settings)
    #                 total_events_array.append(total_events)
    #
    #             total_events_array_copy = np.array(total_events_array)
    #             total_events_array_new = []
    #             for l in range(len(total_events_array)):
    #                 total_events_array_new.append(np.sum(total_events_array_copy[l][-1]))
    #             f_y_i_j = np.array(total_events_array_new)
    #             del total_events_array_copy, total_events_array_new
    #             # f_y_i_j = np.sum(np.sum(np.array(total_events_array), axis=2), axis=1)
    #
    #             del y_i_j
    #
    #             V_j = np.sum(f_z_i * f_y_i_j - E ** 2) / self.NSim
    #
    #             # Calculate Sobol's main effect sensitivity index for each input
    #             s_j = np.abs(
    #                 V_j / V)  # here I take the absolute value (otherwise some of them are negative, negative effect? - not possible)
    #             s_j_array.append(s_j)
    #
    #             T_j = np.sum((f_z_i_ - f_y_i_j) ** 2) / (2 * self.NSim)
    #
    #             # Calculate Sobol's total effect sensitivity index for each input
    #             s_j_t = np.abs(T_j / V)  # same as above for the absolute value
    #             s_j_t_array.append(s_j_t)
    #
    #         result1 = (s_j_array, s_j_t_array)
    #
    #         print("Sensitivity analysis (Sobol global) result (paramater's contribution to output):")
    #         for i in range(len(self.CAPSbounds)):
    #             print(self.OptimizeKeys[i] + ":", s_j_array[i], )
    #
    #         print("Sum of main effect sensitivity indexes:", np.sum(result1[0]), )
    #         print("Sum of total effect sensitivity indexes:", np.sum(result1[1]), )
    #
    #     elif method == 'hoffmann_gardener':
    #
    #         """
    #             Hoffmann and Gardner (1983) calculate the output percent difference when varying one input parameter
    #             from its minimum value to its maximum value. It was introduced to account for all possible values when
    #             determining parameter sensitivity.
    #         """
    #
    #         # self.seed_settings = np.copy(self.seed_settings_optimal)
    #
    #         min_seed_settings = dict.copy(self.seed_settings_optimal)
    #         max_seed_settings = dict.copy(min_seed_settings)
    #         for j in range(len(self.CAPSbounds)):
    #             min_seed_settings[self.OptimizeKeys[j]] = np.min(params_distr[j])
    #             max_seed_settings[self.OptimizeKeys[j]] = np.max(params_distr[j])
    #
    #         # min values
    #         Y_min = []
    #         Y_max = []
    #         s_j = []
    #         for j in range(len(self.CAPSbounds)):
    #             self.seed_settings = dict.copy(self.seed_settings_optimal)
    #             self.seed_settings[self.OptimizeKeys[j]] = min_seed_settings[self.OptimizeKeys[j]]
    #             total_events = CAPS.Forecast_BinnedRate_only_1D(self.TrainingData.time.values[1:], forecast_edges[0],
    #                                                             self.max_pressure_past, self.working_xyz_past,
    #                                                             self.cells_volume_past, self.seed_settings)
    #             Y_min.append(np.sum(total_events[-1, :]))
    #             self.seed_settings = dict.copy(self.seed_settings_optimal)
    #             self.seed_settings[self.OptimizeKeys[j]] = max_seed_settings[self.OptimizeKeys[j]]
    #             total_events = CAPS.Forecast_BinnedRate_only_1D(self.TrainingData.time.values[1:], forecast_edges[0],
    #                                                             self.max_pressure_past, self.working_xyz_past,
    #                                                             self.cells_volume_past, self.seed_settings)
    #             Y_max.append(np.sum(total_events[-1, :]))
    #             s_j.append(abs((Y_max[-1] - Y_min[-1]) / (Y_max[-1])))
    #
    #         print("Sensitivity analysis (Hoffmann & Gardener, 1983) result (paramater's contribution to output):")
    #         for j in range(len(self.CAPSbounds)):
    #             print(self.OptimizeKeys[j] + ":", s_j[j], )
    #
    #         result1 = s_j
    #
    #     else:
    #         """
    #             Local sensitivity analysis: partial derivatives on parameters (very small changes):
    #
    #             Differential sensitivity analysis is based on the partial differentiation of the aggregated model.
    #             A small change in one parameter's value with respect to their calibrated value is tested while the other
    #             parameters are set equal to their calibrated value. The corresponding change in the model's output is evaluated
    #             for that parameter (partial derivative of the dependent variable (input) with respect to the independent variable (output).
    #             The derivatives are then multiplied by the ratio of the parameter value to the model result (normalized by the mean),
    #             to normalize the results by removing the effects of units.
    #         """
    #         z_i = np.array(z_i).T
    #
    #         f_z_i_copy = np.array(f_z_i)
    #         f_z_i_new = []
    #         for l in range(len(f_z_i)):
    #             f_z_i_new.append(np.sum(f_z_i_copy[l][-1]))
    #         f_z_i = np.array(f_z_i_new)
    #         del f_z_i_copy, f_z_i_new
    #         # f_z_i = np.sum(np.sum(np.array(f_z_i), axis=2), axis=1)
    #
    #         z_i_ = z_i + (0.00001 * z_i)  # changing all parameter values by 1/10 of 1% (very small change)
    #
    #         total_events_array_global = []
    #         for j in range(len(self.CAPSbounds)):
    #             z_i_copy = np.copy(z_i)
    #             # replace the added derivative change for one parameter
    #             z_i_copy[j] = z_i_[j]
    #             total_events_array = []
    #             for i in range(len(z_i[0])):
    #                 for l in range(len(self.CAPSbounds)):
    #                     self.seed_settings[self.OptimizeKeys[l]] = z_i_copy[l, i]
    #                 total_events = CAPS.Forecast_BinnedRate_only_1D(self.TrainingData.time.values[1:],
    #                                                                 forecast_edges[0],
    #                                                                 self.max_pressure_past, self.working_xyz_past,
    #                                                                 self.cells_volume_past, self.seed_settings)
    #                 total_events_array.append(np.sum(total_events[-1, :]))
    #             total_events_array_global.append(np.array(total_events_array))
    #         f_z_i_ = np.array(total_events_array_global)
    #
    #         s_j = np.zeros_like(z_i)
    #         s_j_norm = np.copy(s_j)
    #         for j in range(len(self.CAPSbounds)):
    #             for i in range(len(s_j[0])):
    #                 s_j[j, i] = np.abs((f_z_i[i] - f_z_i_[j, i]) / (z_i[j, i] - z_i_[j, i]))
    #                 s_j_norm[j, i] = s_j[j, i] * (z_i[j, i] / f_z_i[i])
    #
    #         # set all values below 0 and above 1 to NaN
    #         s_j_norm[s_j_norm < 0] = np.nan
    #         s_j_norm[s_j_norm > 1] = np.nan
    #
    #         print("Sensitivity analysis (local) mean result (paramater's contribution to output):")
    #         for j in range(len(self.CAPSbounds)):
    #             print(self.OptimizeKeys[j] + ":", np.nanmean(s_j_norm[j]), )
    #
    #         result1 = s_j_norm
    #
    #     return result1

    def Calibrate_CAPS(self, forecast_edges):
        self.update_seeds_settings()  # update b-value

        if self.cost_fun == 'LL':
            self.CostFunction = LL_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
                                                    bins=[self.TrainingData.time.values[1:],
                                                          forecast_edges[0]])
        else:
            self.CostFunction = LS_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
                                                    bins=[self.TrainingData.time.values[1:],
                                                          forecast_edges[0]])
            # self.CostFunction = LS_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
            #                                  bins=[self.TrainingData.time.values[1:],
            #                                        forecast_edges])

        # (radial) spatial bins in negative and positive directions
        self.LL_Edges = self.CostFunction.edges_x2
        self.LL_Moments = self.CostFunction.edges_x1  # time moments

        (self.Training_MaxPressure,
         self.Training_R,
         self.Training_Volume) = self.Interpolate_1D(self.LL_Moments, self.LL_Edges, self.CAPSSettings)

        print("Maximum iterations CAPS:", self.maxiter)

        # open list to save the progress of the minimization
        self.dea_vals = []
        self.dea_func = []

        if self.cost_fun == 'LL':
            result = differential_evolution(self.ll_score, self.SEISbounds, polish=True, disp=self.disp,
                                            maxiter=self.maxiter, seed=self.seed, popsize=10)
        else:
            result = differential_evolution(self.ls_score, self.SEISbounds, polish=True, disp=self.disp,
                                            maxiter=self.maxiter, seed=self.seed, popsize=10,
                                            updating='immediate', callback=self.save_iteration_step)
        print("End CAPS differential evolution")

        print('--------------------------------')
        print('Calibrated geomechanical parameters:')
        for i in range(len(result.x)):
            self.seed_settings[self.OptimizeKeys[i]] = result.x[i]
            if (self.logscale[i]):
                self.seed_settings[self.OptimizeKeys[i]
                                   ] = np.power(10., result.x[i])
            print(str(self.OptimizeKeys[i]) + ': '
                  + str(self.seed_settings[self.OptimizeKeys[i]]))
        print('--------------------------------')

        # self.total_events_past_single, self.total_rates_past_single = CAPS.Caps_Model_linear(self.LL_Moments,
        #                                                                                      self.LL_Edges[self.LL_Edges >= 0.],
        #                                                                                      self.Training_R,
        #                                                                                      self.Training_MaxPressure,
        #                                                                                      self.seed_settings,
        #                                                                                      NSim=1,
        #                                                                                      retriggering=self.retriggering)
        #
        # print("sum of total events after calibration", self.total_events_past_single[0,-1], )

        # save optimal parameters for single forecast later
        self.seed_settings_optimal = dict.copy(self.seed_settings)

        print("**************************************************")
        print("Summary of optimized parameters after calibration:")
        for i in range(len(self.SEISbounds)):
            print(self.OptimizeKeys[i] + ': ',
                  self.seed_settings_optimal[self.OptimizeKeys[i]], )
        j_init = len(self.SEISbounds)
        for j in range(len(self.FlowBounds)):
            print(self.OptimizeKeys[j_init + j] + ': ',
                  self.seed_settings_optimal[self.OptimizeKeys[j_init + j]], )
        print("**************************************************")

        self.total_events_past_grid, self.total_rates_past_grid = CAPS.Caps_Model_linear(self.LL_Moments,
                                                                                         self.LL_Edges[self.LL_Edges >= 0.],
                                                                                         self.Training_R,
                                                                                         self.Training_MaxPressure,
                                                                                         self.seed_settings,
                                                                                         NSim=self.NSim,
                                                                                         retriggering=self.retriggering)

        total_events_median = np.round(
            np.median(self.total_events_past_grid, axis=0)[-1])

        print("sum of total events after calibration (median from grid)",
              total_events_median, )

        # if total_events_median==0:
        #     total_events_median_masked = self.mask_for_median(self.total_events_past_grid[0])
        #
        #     print("sum of total events after calibration and mask (median from grid)", total_events_median_masked[-1], )

        return

    def Calibrate_SEED(self, forecast_edges):
        self.update_seeds_settings()  # update b-value

        if self.cost_fun == 'LL':
            self.CostFunction = LL_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
                                                    bins=[self.TrainingData.time.values, np.linspace(0, max(self.training_catalog[self.Xs].values), 10)[:, 0]])
        else:
            self.CostFunction = LS_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
                                                    bins=[self.TrainingData.time.values, np.linspace(0, max(self.training_catalog[self.Xs].values), 10)[:, 0]])

        self.LL_Edges = self.CostFunction.edges_x2
        self.LL_Moments = self.CostFunction.edges_x1

        # 1D pressure interpolation (in time only)
        pres_sol = (self.flow_model.hidden_mechanism
                    - self.TrainingData[self.Columns[-1]].values[0]) / (1.e6)
        self.pres_t_fit = np.array([np.interp(
            self.LL_Moments[1:], self.flow_model.Q[:, 0], ps_row) for ps_row in pres_sol.T])
        # OR:
        # Tind = np.argmin(np.abs(self.flow_model.Q[:,0] - self.LL_Moments[1:][:, np.newaxis]), axis=1)
        # self.pres_t_fit = (pres_sol[Tind,:]).T

        # Decide for the seed model version (implemented since Poster FAM23)
        if self.seed_model_version == '1':
            self.Nreal = 100  # self.NSim
        elif self.seed_model_version == '2':
            self.Nreal = 1
        else:
            self.Nreal = self.NSim

        print("Maximum iterations SEED:", self.maxiter)

        # open list to save the progress of the minimization
        self.dea_vals = []
        self.dea_func = []

        if self.cost_fun == 'LL':
            result = differential_evolution(self.ll_score_seed, self.SEISbounds, polish=True, disp=self.disp,
                                            maxiter=self.maxiter, seed=self.seed, popsize=10)
        else:
            result = differential_evolution(self.ls_score_seed, self.SEISbounds, polish=True,
                                            disp=self.disp, maxiter=self.maxiter, seed=self.seed, popsize=10,
                                            updating='immediate', atol=1e-3, callback=self.save_iteration_step)
        print("End SEED differential evolution")

        print('------------------------------------')
        print('Calibrated geomechanical parameters:')
        for i in range(len(result.x)):
            self.seed_settings[self.OptimizeKeys[i]] = result.x[i]
            if (self.logscale[i]):
                self.seed_settings[self.OptimizeKeys[i]
                                   ] = np.power(10., result.x[i])
            print(str(self.OptimizeKeys[i]) + ': '
                  + str(self.seed_settings[self.OptimizeKeys[i]]))
        print('------------------------------------')

        # save optimal parameters for forecast later
        self.seed_settings_optimal = dict.copy(self.seed_settings)

        print("**************************************************")
        print("Summary of optimized parameters after calibration:")
        for i in range(len(self.SEISbounds)):
            print(self.OptimizeKeys[i] + ': ',
                  self.seed_settings_optimal[self.OptimizeKeys[i]], )
        j_init = len(self.SEISbounds)
        for j in range(len(self.FlowBounds)):
            print(self.OptimizeKeys[j_init + j] + ': ',
                  self.seed_settings_optimal[self.OptimizeKeys[j_init + j]], )
        print("**************************************************")

        # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
        SeedsSettings = nb.typed.List()
        for j in range(len(self.SEEDmodelKeys)):
            SeedsSettings.append(
                float(self.seed_settings[self.SEEDmodelKeys[j]]))

        self.training_catalog, self.total_events_past_grid, self.total_rates_past_grid = SEED.Seed_Model_linear(self.LL_Moments[1:],
                                                                                                                self.LL_Edges[
                                                                                                                    self.LL_Edges >= 0.],
                                                                                                                self.flow_model.r,
                                                                                                                self.flow_model.Q[:, 0],
                                                                                                                self.pres_t_fit,
                                                                                                                SeedsSettings,
                                                                                                                self.NSim,
                                                                                                                self.retriggering)

        total_events_median = np.round(
            np.median(self.total_events_past_grid, axis=0)[-1])

        print("sum of total events after calibration (median from grid)",
              total_events_median, )

        return

    def _fit_sequential(self, HydraulicLogs_coarse_pd, catalog_incomplete_df, forecast_edges, deep=True):
        self.TrainingData = HydraulicLogs_coarse_pd.copy(deep)
        self.training_catalog = catalog_incomplete_df.copy(deep)

        # filepath = '/home/victor/Desktop/pyatls/Data/Bedretto/FlowModels/'

        if self.flw_model == 'point_source':
            print("HM0: calibrating flow model (point source)....")
            self.flow_model.fit(self.TrainingData[self.Columns[::2]].values)
            # or:
            # with open(filepath + 'HM0_point_source.pkl', 'rb') as f:
            #     self.flow_model = pickle.load(f)
            # self.flow_model = self.flow_model[0]
        elif self.flw_model == 'line_source':
            print("HM0: calibrating flow model (line source)....")
            self.flow_model.fit(self.TrainingData[self.Columns[::2]].values)
            # or:
            # with open(filepath + 'HM0_line_source.pkl', 'rb') as f:
            #     self.flow_model = pickle.load(f)
            # self.flow_model = self.flow_model[0]
        elif self.flw_model == 'num_const_perm':
            print("HM0: calibrating flow model (numerical linear)....")
            self.flow_model.fit(
                self.TrainingData[self.Columns[:2]].values, self.TrainingData[self.Columns[::2]].values)
            # or:
            # with open(filepath + 'HM0_numerical_linear.pkl', 'rb') as f:
            #     self.flow_model = pickle.load(f)
            # self.flow_model = self.flow_model[0]
        elif self.flw_model == 'num_var_perm':
            print("HM1: calibrating flow model (numerical non-linear)....")
            self.flow_model.fit(
                self.TrainingData[self.Columns[:2]].values, self.TrainingData[self.Columns[::2]].values)
            # or:
            # with open(filepath + 'HM1_2.pkl', 'rb') as f:
            #     self.flow_model = pickle.load(f)
            # self.flow_model = self.flow_model[0]
        else:
            print("HM1: calibrating flow model (numerical non-linear with FV)....")
            self.flow_model.fit(
                self.TrainingData[self.Columns[:2]].values, self.TrainingData[self.Columns[::2]].values)

        print('--------------------------------')
        print('Calibrated hydraulic parameters:')
        for i in range(len(self.flow_model.params)):
            self.seed_settings[self.OptimizeKeys[i
                                                 - len(self.flow_model.params)]] = self.flow_model.params[i]
            print(str(self.OptimizeKeys[i - len(self.flow_model.params)]
                      ) + ': ' + str(self.flow_model.params[i]))
        print('--------------------------------')

        self.timesteps = self.TrainingData.time.values
        inipres = self.TrainingData[self.Columns[-1]].values[0]

        if self.seis_model == 'CAPS':
            print("HM0/1: calibrating CAPS....")
            self.PressureFunction = self.create_interpolator(
                self.timesteps, (self.flow_model.hidden_mechanism - inipres) / 1.e6)
            self.Calibrate_CAPS(forecast_edges)
        else:
            print("HM0/1: calibrating SEED....")
            self.Calibrate_SEED(forecast_edges)

        return

    # def _fit_joint(self, HydraulicLogs_coarse_pd, catalog_incomplete_df, forecast_edges, deep=True):
    #     self.TrainingData = HydraulicLogs_coarse_pd.copy(deep)
    #     self.training_catalog = catalog_incomplete_df.copy(deep)
    #
    #     self.update_seeds_settings()
    #
    #     if self.cost_fun == 'LL':
    #         self.SeisCostFunction = LL_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
    #                                                     bins=[self.TrainingData.time.values[1:],
    #                                                           forecast_edges[0]])
    #     else:
    #         self.SeisCostFunction = LS_cost_function_1D(np.array(self.training_catalog[np.append(self.ts, self.Xs)]),
    #                                                     bins=[self.TrainingData.time.values[1:],
    #                                                           forecast_edges[0]])
    #
    #     self.LL_Edges = self.SeisCostFunction.edges_x2
    #     self.LL_Moments = self.SeisCostFunction.edges_x1
    #
    #     # We need to join the CAPS and the Flow bounds stated previously:
    #     self.ParamBounds = self.CAPSbounds + self.FlowBounds
    #
    #     print("Maximum iterations CAPS:", self.maxiter)
    #     self.CAPS_iter_results = []
    #     if self.cost_fun=='LL':
    #         result = differential_evolution(self.ll_score_joint, self.ParamBounds,polish=True,
    #                                     disp=self.disp, maxiter=self.maxiter, seed=self.seed,popsize=10, callback=self.caps_iteration)
    #     else:
    #         result = differential_evolution(self.ls_score_joint, self.ParamBounds,polish=True,
    #                                         disp=self.disp, maxiter=100, seed=self.seed,popsize=10, callback=self.caps_iteration)
    #
    #     print("End CAPS differential evolution")
    #
    #     for i in range(len(result.x)):
    #         self.seed_settings[self.OptimizeKeys[i]] = result.x[i]
    #         if (self.logscale[i]):
    #             self.seed_settings[self.OptimizeKeys[i]] = np.power(10., result.x[i])
    #
    #     self.flow_model.params = self.seed_settings[self.OptimizeKeys[4]], self.seed_settings[self.OptimizeKeys[5]]
    #     #self.flow_model.params = self.seed_settings[self.OptimizeKeys[4]]
    #
    #     self.flow_model.hidden_mechanism = self.flow_model.predict()
    #
    #     self.timesteps = self.TrainingData.time.values
    #     inipres=self.TrainingData[self.Columns[-1]].values[0]
    #     self.PressureFunction = self.create_interpolator(self.timesteps,(self.flow_model.hidden_mechanism-inipres)/1.e6)
    #
    #     (max_pressure_past,
    #      working_xyz_past,
    #      cells_volume_past) = self.Interpolate_1D(self.TrainingData.time.values[1:], forecast_edges[0],
    #                                               self.ExSolSettings, self.CAPSSettings)
    #
    #     self.total_events_past_single = CAPS.Forecast_BinnedRate_only_1D(
    #         self.TrainingData.time.values[1:], forecast_edges[0], max_pressure_past, working_xyz_past,
    #         cells_volume_past, self.seed_settings)
    #
    #     print('Poisson Density CAPS:', self.seed_settings["PoissonDensity"])
    #     print("sum of total events after calibration", np.sum(self.total_events_past_single[-1, :]), )
    #
    #     # looking for index in event past corresponsing to time of last event
    #     index = np.where(self.TrainingData.time.values[1:] <= self.training_catalog.ts.values[-1])
    #     NewPoisson = (self.seed_settings['PoissonDensity']
    #                   * len(self.training_catalog) / (
    #                       np.sum(self.total_events_past_single[index[0][-1] + 1, :])))  # noqa
    #
    #     # Only adapting the Poisson Density if it is  a finite number and if the predicted cumulative events
    #     # are within a specific range from the observed ones:
    #     events_ratio = self.SeisCostFunction.y[-1]/(np.sum(self.total_events_past_single[index[0][-1]+1, :]))
    #     if np.isfinite(NewPoisson):
    #         if 0.85 <= events_ratio <= 1.15:
    #             self.seed_settings['PoissonDensity'] = NewPoisson
    #             print('Poisson density CAPS within the range. Corrected.')
    #         else:
    #             print('Poisson density CAPS is finite but not within the range. Not corrected.')
    #     else:
    #         print('Poisson Density CAPS is not finite. Not corrected.')
    #
    #     print('Poisson Density CAPS:', self.seed_settings["PoissonDensity"])
    #
    #     self.total_events_past_single = CAPS.Forecast_BinnedRate_only_1D(
    #         self.TrainingData.time.values[1:], forecast_edges[0], max_pressure_past, working_xyz_past,
    #         cells_volume_past, self.seed_settings)
    #
    #     # save optimal parameters for single forecast later
    #     self.seed_settings_optimal = self.seed_settings
    #     # save maximum pressure past for later
    #     self.max_pressure_past = max_pressure_past
    #     self.working_xyz_past = working_xyz_past
    #     self.cells_volume_past = cells_volume_past
    #
    #     print("sum of total events after Re-calibration (calibrated params)", np.sum(self.total_events_past_single[-1, :]), )
    #
    #     print("Optimized parameters after calibration:")
    #     for i in range(len(self.CAPSbounds)):
    #         print(self.OptimizeKeys[i]+":", self.seed_settings[self.OptimizeKeys[i]], )
    #
    #     j_init = len(self.CAPSbounds)
    #     for j in range(len(self.FlowBounds)):
    #         print(self.OptimizeKeys[j_init+j]+":", self.seed_settings[self.OptimizeKeys[j_init+j]], )
    #
    #     #---------------------------------------------------------------------------------------------------------------
    #
    #     # get uncertainty range for parameters
    #     # par_grd_1 = self.Params_flowmodel()
    #
    #     cal_val = []
    #     for i in range(len(self.CAPSbounds)):
    #         cal_val.append(self.seed_settings[self.OptimizeKeys[i]])
    #     self.cal_val = np.array(cal_val)
    #
    #     # result = differential_evolution(self.ls_params_score, [(90, 100)], polish=True, disp=self.disp,
    #     #                                 maxiter=self.maxiter, seed=self.seed, popsize=10, callback=self.caps_iteration)
    #
    #     # percent = result.x
    #     percent = 10
    #
    #     # print("1sigma for normal trunc. distr. of params: " + str(percent[0]) + " divided by 100 times calib. parameter.")
    #
    #     par_grd_2 = []
    #     for p in range(len(self.CAPSbounds)):
    #         # par_distr = self.Params_CAPS(forecast_edges, max_pressure_past, working_xyz_past, cells_volume_past, p)
    #         par_distr = self.Params_CAPS_2(percent, p)
    #         # par_grd_2.append((par_distr.T)[0])
    #         par_grd_2.append(par_distr)
    #
    #     self.par_grd = np.array(par_grd_2)
    #     # self.par_grd = np.concatenate((par_grd_2, par_grd_1))
    #     # self.par_grd = np.vstack((par_grd_2, par_grd_1))
    #
    #     # Plotting distributions
    #     #ATLS.PlotParamDist(self.par_grd, self.cal_val, seis_model='CAPS', filename='Bedretto_params_distr.png')
    #
    #     if 1 <= self.workers <= mp.cpu_count():
    #         nr_cpu = self.workers
    #     else:
    #         print('Workers must be between 1 and 16 or -1 to indicate max. nr. of cores, which is 16')
    #         print('Default: using all 16 cores')
    #         nr_cpu = mp.cpu_count()
    #
    #     self.nr_cpu = nr_cpu
    #
    #     pool = mp.Pool(nr_cpu)
    #     result = pool.starmap(self.Parallel_CAPS_fitting, [(nr_sim, self.TrainingData.time.values[1:], forecast_edges, self.par_grd) for nr_sim in range(self.NSim)])
    #     pool.close()
    #
    #     # total_events = [result[a][0] for a in range(len(result))]
    #     # press_fit = [result[a][1] for a in range(len(result))]
    #
    #     self.total_events_past_grd = (np.array(result)).T
    #     # self.press_fit = (np.array(press_fit)).T
    #
    #     total_events_median = np.round(np.sum(np.median(self.total_events_past_grd, axis=2), axis=0))[-1]
    #
    #     print("sum of total events after Re-calibration (median from grid)", total_events_median, )
    #
    #     if total_events_median == 0:
    #         total_events_median_masked = self.mask_for_median(self.total_events_past_grd)
    #
    #         print("sum of total events after Re-calibration and mask (median from grid)",
    #               total_events_median_masked[-1], )
    #
    #     return

    def fit(self, HydraulicLogs_coarse_pd, catalog_incomplete_df, forecast_edges, deep=True):

        if self.inv_mod == 'joint':
            print("Model not available at the moment ... using sequential.")
            self._fit_sequential(HydraulicLogs_coarse_pd,
                                 catalog_incomplete_df, forecast_edges, deep)
            # self._fit_joint(HydraulicLogs_coarse_pd, catalog_incomplete_df, forecast_edges, deep)
        else:
            self._fit_sequential(HydraulicLogs_coarse_pd,
                                 catalog_incomplete_df, forecast_edges, deep)

    def predict(self, PlannedStrategy, Forecasts_Moments, Forecasts_Edges):

        # CAPS MODEL #

        if self.seis_model == 'CAPS':
            self.seed_settings = self.seed_settings_optimal

            (self.max_pressure_past,
             self.Forecast_R,
             self.Forecast_Volume) = self.Interpolate_1D(Forecasts_Moments, Forecasts_Edges[0], self.CAPSSettings)

            self.press_prediction = self.flow_model.predict(
                np.array(PlannedStrategy[["time", "flow"]].values))

            self.timesteps = PlannedStrategy.time.values
            inipres = self.TrainingData[self.Columns[-1]].values[0]
            self.PressureFunction = self.create_interpolator(
                self.timesteps, (self.press_prediction - inipres) / 1.e6)

            (self.Forecast_MaxPressure,
             self.Forecast_R,
             self.Forecast_Volume) = self.Interpolate_1D(Forecasts_Moments, Forecasts_Edges[0], self.CAPSSettings)

            max_pressure = np.maximum(
                self.max_pressure_past, self.Forecast_MaxPressure)

            print('Starting CAPS forecast')

            self.total_events_forecast_grid, self.total_rates_forecast_grid = CAPS.Caps_Model_linear(Forecasts_Moments,
                                                                                                     Forecasts_Edges[0][
                                                                                                         Forecasts_Edges[0] >= 0.],
                                                                                                     self.Forecast_R,
                                                                                                     max_pressure,
                                                                                                     self.seed_settings,
                                                                                                     NSim=self.NSim,
                                                                                                     retriggering=self.retriggering)

            print('Ending CAPS forecast')

            # This is a "trick" and needs to be deleted after the FAM23

            # if np.median(self.total_events_forecast_grid, axis=0)[0]==0:
            #     if self.flow_model.Q[0, 1] <= 1.e-9:
            #         # repair events matrix
            #         last_col = self.total_events_past_grid[:, -1]
            #         total_events_forecast_grid_copy = np.zeros_like((self.total_events_forecast_grid))
            #         for i in range(len(self.total_events_forecast_grid[0])):
            #             total_events_forecast_grid_copy[:, i] = self.total_events_forecast_grid[:, i] + last_col
            #         del self.total_events_forecast_grid
            #         self.total_events_forecast_grid = total_events_forecast_grid_copy
            #         del total_events_forecast_grid_copy
            #         # repair rates matrix
            #         # self.total_rates_forecast_grid[:, 0] = self.total_rates_past_grid[:,-1]
            #         # Replace the first time moment for both events and rates by the last one from training
            #         self.total_events_forecast_grid[:, 0] = self.total_events_forecast_grid[:, 1]
            #         diff_col = abs(self.total_events_past_grid[:, -1] - self.total_events_forecast_grid[:, 0])
            #         total_events_forecast_grid_copy = np.zeros_like((self.total_events_forecast_grid))
            #         for i in range(len(self.total_events_forecast_grid[0])):
            #             total_events_forecast_grid_copy[:, i] = self.total_events_forecast_grid[:, i] - diff_col
            #         del self.total_events_forecast_grid
            #         self.total_events_forecast_grid = total_events_forecast_grid_copy
            #     else:
            #         # Replace the first time moment for both events and rates by the last one from training
            #         self.total_events_forecast_grid[:, 0] = self.total_events_past_grid[:, -1]

            return

    # SEED MODEL #

        else:
            self.seed_settings = self.seed_settings_optimal

            self.timesteps = PlannedStrategy.time.values
            self.press_prediction = self.flow_model.predict(
                np.array(PlannedStrategy[["time", "flow"]].values))

            # 1D pressure interpolation (in time only)
            pres_sol = (self.press_prediction
                        - self.TrainingData[self.Columns[-1]].values[0]) / (1.e6)
            self.pres_t_forecast = np.array([np.interp(Forecasts_Moments, PlannedStrategy['time'].values, ps_row)
                                             for ps_row in pres_sol.T])
            # OR:
            # Tind = np.argmin(np.abs(PlannedStrategy['time'].values - Forecasts_Moments[:, np.newaxis]), axis=1)
            # self.pres_t_fit = (pres_sol[Tind,:]).T

            # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
            SeedsSettings = nb.typed.List()
            for j in range(len(self.SEEDmodelKeys)):
                SeedsSettings.append(
                    float(self.seed_settings[self.SEEDmodelKeys[j]]))

            # self.synthetic_catalog, self.total_events_forecast_grid, self.total_rates_forecast_grid = SEED.Seed_Model_linear(Forecasts_Moments,
            #                                                                                                                  Forecasts_Edges[0][Forecasts_Edges[0] >= 0.],
            #                                                                                                                  self.flow_model.r,
            #                                                                                                                  PlannedStrategy['time'].values,
            #                                                                                                                  self.pres_t_forecast,
            #                                                                                                                  SeedsSettings,
            #                                                                                                                  self.NSim,
            #                                                                                                                  self.retriggering)

            self.synthetic_catalog, self.total_events_grid, self.total_rates_grid = SEED.Seed_Model_linear(np.concatenate((self.LL_Moments[1:], Forecasts_Moments[1:]), axis=0),
                                                                                                           Forecasts_Edges[0][
                                                                                                               Forecasts_Edges[0] >= 0.],
                                                                                                           self.flow_model.r,
                                                                                                           np.concatenate(
                                                                                                               (self.flow_model.Q[:, 0], PlannedStrategy['time'].values[1:]), axis=0),
                                                                                                           np.concatenate(
                                                                                                               (self.pres_t_fit, self.pres_t_forecast[:, 1:]), axis=1),
                                                                                                           SeedsSettings,
                                                                                                           self.NSim,
                                                                                                           self.retriggering)
            self.total_events_past_grid = self.total_events_grid[:, :len(
                self.LL_Moments[1:])]
            self.total_events_forecast_grid = self.total_events_grid[:, (len(
                self.LL_Moments[1:]) - 1):]

            # self.total_rates_forecast_grid = self.total_rates_grid[:, (len(self.total_rates_past_grid[0])-1):-1]

            # OR:

            # Take the rates of the forecast and sum up to the last events of the training
            # total_events = np.zeros_like(self.total_events_forecast_grid)
            # total_events[:,0] = self.total_events_past_grid[:,-1]
            # for i in range(1, len(total_events[0])):
            #     total_events[:,i] = total_events[:,i-1] + self.total_rates_forecast_grid[:,i-1]
            #
            # self.total_events_forecast_grid = total_events

            return

    def forecast_only(self, PlannedStrategy, Forecasts_Moments, Forecasts_Edges):

        # CAPS MODEL #

        if self.seis_model == 'CAPS':

            self.press_prediction = self.flow_model.predict(
                np.array(PlannedStrategy[["time", "flow"]].values))

            self.timesteps = PlannedStrategy.time.values
            inipres = self.TrainingData[self.Columns[-1]].values[0]
            self.PressureFunction = self.create_interpolator(
                self.timesteps, (self.press_prediction - inipres) / 1.e6)

            (self.Forecast_MaxPressure,
             self.Forecast_R,
             self.Forecast_Volume) = self.Interpolate_1D(Forecasts_Moments, Forecasts_Edges[0], self.CAPSSettings)

            print('Starting CAPS forecast')

            self.total_events_forecast_grid, self.total_rates_forecast_grid = CAPS.Caps_Model_linear(Forecasts_Moments,
                                                                                                     Forecasts_Edges[0][
                                                                                                         Forecasts_Edges[0] >= 0.],
                                                                                                     self.Forecast_R,
                                                                                                     self.Forecast_MaxPressure,
                                                                                                     self.seed_settings,
                                                                                                     NSim=self.NSim,
                                                                                                     retriggering=self.retriggering)

            print('Ending CAPS forecast')

            return

    # SEED MODEL #

        else:

            self.timesteps = PlannedStrategy.time.values
            self.press_prediction = self.flow_model.predict(
                np.array(PlannedStrategy[["time", "flow"]].values))

            # 1D pressure interpolation (in time only)
            pres_sol = (self.press_prediction
                        - self.TrainingData[self.Columns[-1]].values[0]) / (1.e6)
            self.pres_t_forecast = np.array([np.interp(
                Forecasts_Moments, PlannedStrategy['time'].values, ps_row) for ps_row in pres_sol.T])

            # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
            SeedsSettings = nb.typed.List()
            for j in range(len(self.SEEDmodelKeys)):
                SeedsSettings.append(
                    float(self.seed_settings[self.SEEDmodelKeys[j]]))

            self.synthetic_catalog, self.total_events_grid, self.total_rates_grid = SEED.Seed_Model_linear(Forecasts_Moments,
                                                                                                           Forecasts_Edges[0][
                                                                                                               Forecasts_Edges[0] >= 0.],
                                                                                                           self.flow_model.r,
                                                                                                           PlannedStrategy['time'].values,
                                                                                                           self.pres_t_forecast,
                                                                                                           SeedsSettings,
                                                                                                           self.NSim,
                                                                                                           self.retriggering)

            # self.total_events_past_grid = self.total_events_grid[:, :len(self.LL_Moments[1:])]
            # self.total_events_forecast_grid = self.total_events_grid[:, (len(self.LL_Moments[1:])-1):]

            return
