import copy
import logging
from datetime import timedelta

import numba as nb
import numpy as np
from hermes_model import ModelInput
from hermes_model.logging import set_up_logging
from hydws.parser import BoreholeHydraulics
from scipy import stats
from scipy.interpolate import interp1d
from scipy.optimize import differential_evolution
from seismostats import Catalog, ForecastGRRateGrid
from seismostats.utils.coordinates import CoordinateTransformer
from seismostats.analysis.bvalue import estimate_b
from seismostats.analysis.estimate_mc import estimate_mc_maxc
from seismostats.utils.binning import bin_to_precision
from shapely import wkt
from sklearn.linear_model import Lasso

from hm1d.FlowModel import (FlowModel_an1, FlowModel_an2, FlowModel_kchange,
                            FlowModel_num)
from hm1d.SeismicityModel import Caps_Model_linear as CAPS
from hm1d.SeismicityModel import Seed_Model_linear as SEED
from hm1d.utils import (Calibrate_Observed_GR, Interpolate_1D,
                        LS_cost_function_1D, b_value_Lasso_calibration_1D,
                        create_interpolator, daterange, forecast_space_bins)


class HM1D():
    """ InducedSFM Class for Hydro-mechanical models in 1D

        Available variables after successfull initialization:

        Input Hydraulic and Seismic data is available as DataFrames as:
        self.hydraulic_df,
        self.scenario_df,
        self.seismicity_df

        Variables from input dict are also available as class properties:
        self.model_parameters
        self.forecast_start
        self.forecast_end
        self.epoch_duration
        self.proj_string

        Additionally are available:
        self.forecast_moments
        self.section_metadata
        self.observation_start
        self.observation_end

        A logger is available with self._logger.* (info, warning, error, ...)

    """

    def __init__(self, model_input: ModelInput, **kwargs):
        """
        Gets called after the aforementioned variables are instantiated, eg.
        model setup can live in here.
        """
        # set up logger
        set_up_logging(__name__)
        self._logger = logging.getLogger(__name__)
        self._logger.info('Model Initialized')

        self.model_input = model_input
        self.parameters = \
            model_input.model_parameters | model_input.model_settings
        self.proj_string = self.parameters['local_proj_string']
        self.delta_tau = self.parameters['Delta_tau']
        self.forecastDX = self.parameters.get('forecast_dx',20)

        self.column_flow = "topflow"
        self.column_pressure = "toppressure"
        self.Columns = ["time", "topflow", "toppressure"]

        self.seed_settings = self.parameters['SeisSettings']
        self.seis_model = self.seed_settings['seis_model']
        listOfStrings = ['SEED', 'CAPS']
        if self.seis_model not in listOfStrings:
            self._logger.warning(
                f'Seismic model "{self.seis_model}" not found. '
                f'Possible values are {listOfStrings} '
                f'Using default mode (CAPS).')
            self.seis_model = 'CAPS'

        self.NSim = self.seed_settings['NSim']
        self.retriggering = self.seed_settings.get('retriggering', False)
        self.seed_model_version = self.seed_settings.get(
            'seed_model_version', 2)
        self.seis_bounds = self.seed_settings['SeisBounds']
        # upper b-value limit
        #self.maxbval_bounds = [(1., self.seed_settings["b_max"])]
        self.dM = self.seed_settings['dM']
        self.Mc = self.seed_settings.get('Mc', None)
        self.Mc_corr = self.seed_settings.get('mc_corr', 0)
        self.seed_settings['criticality'] = self.seed_settings.get('criticality', 0)
        self.seed_settings['b_vs_depth_range'] = self.seed_settings.get('b_vs_depth_range', False)

        self.maxiter = self.parameters.get('max_iter_DE', 200)
        self.disp = self.parameters.get('DE_verbose', True)
        self.seed = 42
        # for callback
        self.best_x = None
        self.minf = np.inf

        # self.cost_fun = self.seed_settings['cost_fun']
        # listOfStrings = ['LL' , 'LS']
        # if self.cost_fun not in listOfStrings :
        #    self._logger.warning("Cost function '"+self.cost_fun+"' not found. Possible values are:"+str(listOfStrings)+". Using default mode (LS).")
        #    self.cost_fun='LS'

        self.CostFunction = None
        self.LL_Edges = None
        self.LL_Moments = None
        self.Training_MaxPressure = None
        self.Training_XYZ = None
        self.Training_Volume = None

        # input for the flow models
        self.flow_settings = self.parameters['FlowSettings']
        self.flow_bounds = self.flow_settings['FlowBounds']
        self.max_grid_ext = self.flow_settings.get('max_grid_ext', 300)
        self.layer_thick = self.flow_settings.get('layer_thick', 100)
        self.cross_area = self.flow_settings.get('cross_area', 94)

        self.flw_model = self.flow_settings['flow_model']
        listOfStrings = ['point_source', 'line_source',
                         'num_const_perm', 'num_var_perm']
        if self.flw_model not in listOfStrings:
            self._logger.warning(
                f'Fluid model "{self.flw_model}" not found. '
                f'Possible values are {listOfStrings} '
                f'Using default mode (CAPS).')
            self.seis_model = 'line_source'

        # DATA INITIALIZATION

        section_id = self.parameters['well_section_id']

        self.bounding_polygon = wkt.loads(model_input.bounding_polygon)
        self.depth_min = model_input.depth_min
        self.depth_max = model_input.depth_max

        self.forecast_start = model_input.forecast_start
        self.forecast_end = model_input.forecast_end

        self.hydraulics = BoreholeHydraulics(
            model_input.injection_observation[0])

        self.scenario = BoreholeHydraulics(model_input.injection_plan[0])

        self.seismicity = Catalog.from_quakeml(
            model_input.seismicity_observation,
            include_quality=True)

        # get dataframes from HYDSJONParser
        self.hydraulic_df = self.hydraulics[section_id].hydraulics
        self.hydraulic_df = self.hydraulic_df[~self.hydraulic_df.index.duplicated(
            keep='first')]
        self.hydraulic_df = self.hydraulic_df[
            ~self.hydraulic_df['topflow'].isna()]
        self.hydraulic_df = self.hydraulic_df[
            ~self.hydraulic_df['toppressure'].isna()]
        # self.hydraulic_df['time'] = \
        #     np.array((self.hydraulic_df.index.values
        #               - self.hydraulic_df.index.values[0]).tolist()) / 1.e9
        self.hydraulic_df['time'] = self.hydraulic_df.index.map(
            lambda x: (x - self.forecast_start).total_seconds())

        self.scenario_df = self.scenario[section_id].hydraulics

        self.observation_start = self.hydraulic_df.index.min()
        self.observation_end = np.array(
            self.forecast_start).astype('datetime64')

        self.hydraulic_df = self.hydraulic_df[self.hydraulic_df.index
                                              < self.observation_end]

        self.hydraulic_df = self.hydraulic_df.resample(
            str(self.parameters['Delta_tau']) + 's').mean().interpolate(method='linear')


        self.seismicity = self.seismicity[self.seismicity['time']
                                          < self.forecast_start]
        self.seismicity = self.seismicity[self.seismicity['time']
                                          >= self.observation_start]

        # self.observation_end = self.hydraulic_df.index.max()
        # self.observation_end = min(self.observation_end, self.forecast_start)

        self.seismicity = self.seismicity[~self.seismicity['magnitude'].isna()]
        self.seismicity['magnitude'] = bin_to_precision(
            self.seismicity['magnitude'], self.dM)

        if 'associatedphasecount' in self.seismicity.columns:
            self.seismicity['associatedphasecount'] = \
                self.seismicity['associatedphasecount'].astype(float)
            # is this sensible?
            self.seismicity['associatedphasecount'] = \
                self.seismicity['associatedphasecount'].fillna(0)

            self.seismicity = \
                self.seismicity[self.seismicity['associatedphasecount']
                                >= self.parameters['n_phases']]

        # add ts column to seismicity dataframe
        self.seismicity['ts'] = self.seismicity.time.map(
            lambda x: (x - self.forecast_start).total_seconds())

        self.injection_point = self.parameters['injection_point']

        transformer = CoordinateTransformer(
            self.proj_string,
            ref_easting=self.injection_point[0],
            ref_northing=self.injection_point[1],
            ref_altitude=self.injection_point[2],  # needs to match seismicity
            external_proj=4326,
            ref_proj=4326  # injection point given as WGS84
        )

        self.seismicity['Xs'], self.seismicity['Ys'], self.seismicity['Zs'] = \
            transformer.to_local_coords(self.seismicity['longitude'],
                                        self.seismicity['latitude'],
                                        self.seismicity['depth'])

        self.seismicity['Rs'] = np.sqrt((self.seismicity['Xs']) ** 2
                                        + (self.seismicity['Ys']) ** 2
                                        + (self.seismicity['Zs']) ** 2)

        forecast_bounds = np.array(
            [-np.max(self.seismicity['Rs']), np.max(self.seismicity['Rs'])])
        self.forecast_edges = forecast_space_bins(forecast_bounds, self.forecastDX)

        # get the forecast times
        self.forecast_moments = daterange(self.forecast_start,
                                          self.forecast_end,
                                          self.parameters['epoch_duration'],
                                          self.observation_start)

        #self.scenario_df = self.scenario_df.resample(
        #    str(self.parameters['epoch_duration']) + 's').mean().interpolate(method='linear')

        self.scenario_df = self.scenario_df.resample(
            str(self.parameters['Delta_tau']) + 's').mean().interpolate(method='linear')

        self.scenario_df = self.scenario_df[(self.scenario_df.index <= self.forecast_end)
                                            & (self.scenario_df.index > self.forecast_start)]

        self.scenario_df['time'] = self.scenario_df.index.map(
            lambda x: (x - self.forecast_start).total_seconds())

        sc_all_negative = (self.scenario_df['topflow'] < 0).all()
        sc_all_identical = self.scenario_df['topflow'].nunique() == 1

        if sc_all_negative and sc_all_identical:
            self.scenario_df["topflow"] = (-self.scenario_df["topflow"].values[0]
                                           * np.ones(len(self.scenario_df)) * self.hydraulic_df["topflow"].values[-1])

        sc_all_negative = (self.scenario_df['topflow'] < 0).all()
        sc_all_identical = self.scenario_df['topflow'].nunique() == 1

        if sc_all_negative and sc_all_identical:
            self.scenario_df["topflow"] = (-self.scenario_df["topflow"].values[0]
                                           * np.ones(len(self.scenario_df)) * self.hydraulic_df["topflow"].values[-1])

        self.training_catalog = self.seismicity.copy(deep=True)

        # MODEL INIT

        if self.retriggering == True:
            if self.seed_settings['criticality'] != 0:
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'stressdrop_coeff',
                                     'criticality', 'cohesion_mean', 'friction_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['stressdrop_coeff'][0],
                                    self.seis_bounds['stressdrop_coeff'][1]),
                                   (self.seis_bounds['criticality'][0],
                                    self.seis_bounds['criticality'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False,
                                 False, False, False, False]
            else:
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std', 'stressdrop_coeff',
                                     'cohesion_mean', 'friction_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['stressdrop_coeff'][0],
                                    self.seis_bounds['stressdrop_coeff'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False, False, False, False]
        else:
            if self.seed_settings['criticality'] != 0:
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std',
                                     'friction_mean', 'cohesion_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False, False, False]
            else:
                self.OptimizeKeys = ['sigma1_std', 'sigma3_std',
                                     'cohesion_mean', 'friction_mean', 'seed_density']
                self.SEISbounds = [(self.seis_bounds['sigma1_std'][0], self.seis_bounds['sigma1_std'][1]),
                                   (self.seis_bounds['sigma3_std'][0],
                                    self.seis_bounds['sigma3_std'][1]),
                                   (self.seis_bounds['cohesion_mean'][0],
                                    self.seis_bounds['cohesion_mean'][1]),
                                   (self.seis_bounds['friction_mean'][0],
                                    self.seis_bounds['friction_mean'][1]),
                                   (self.seis_bounds['seed_density'][0], self.seis_bounds['seed_density'][1])]
                self.logscale = [False, False, False, False, False]

        """
        4 different flow models can be choosen:
        -> point_source: analytical point source solution. Two parameters, storage coefficient and permeability, are calibrated.
        -> line_source: Theis' solution. Two parameters, storage coefficient and permeability, are calibrated.
        -> num_const_perm: numerical solution without permeability changes. Storage and permeability are calibrated.
        -> num_var_perm: numerical with variable fracture permeability. Parameters to be calibrated are:
                ['zero_stress_aperture', 
                'residual_aperture', 
                'shearing_aperture_factor', 
                'max_shear_frac_aperture', 
                'shear_scaling_factor', 
                'storage']
        """

        if self.flw_model == 'point_source':
            self.OptimizeKeys = self.OptimizeKeys + \
                ['storage', 'permeability']  # parameters to calibrate
            self.FlowBounds = [(self.flow_bounds['storage'][0], self.flow_bounds['storage'][1]),
                               (self.flow_bounds['permeability'][0], self.flow_bounds['permeability'][1])]
            self.logscale = self.logscale + [True, True]
            # activate the flow model:
            self.flow_model = FlowModel_an1(self.hydraulic_df[self.Columns[:2]].values,
                                            np.geomspace(
                                                0.1, self.max_grid_ext, int(self.max_grid_ext/self.forecastDX), endpoint=True),
                                            self.cross_area,
                                            self.flow_settings['viscosity'],
                                            bounds=self.FlowBounds,
                                            iniguess=(self.flow_settings['storage'], self.flow_settings['permeability']))

        elif self.flw_model == 'line_source':
            self.OptimizeKeys = self.OptimizeKeys + ['storage', 'permeability']
            self.FlowBounds = [(self.flow_bounds['storage'][0], self.flow_bounds['storage'][1]),
                               (self.flow_bounds['permeability'][0], self.flow_bounds['permeability'][1])]
            self.logscale = self.logscale + [True, True]
            # activate the flow model:
            self.flow_model = FlowModel_an2(self.hydraulic_df[self.Columns[:2]].values,
                                            np.geomspace(
                                                0.1, self.max_grid_ext, int(self.max_grid_ext/self.forecastDX), endpoint=True),
                                            self.layer_thick,
                                            self.flow_settings['viscosity'],
                                            bounds=self.FlowBounds,
                                            iniguess=(self.flow_settings['storage'], self.flow_settings['permeability']))

        elif self.flw_model == 'num_const_perm':
            self.OptimizeKeys = self.OptimizeKeys + ['storage', 'permeability']
            self.FlowBounds = [(self.flow_bounds['storage'][0], self.flow_bounds['storage'][1]),
                               (self.flow_bounds['permeability'][0], self.flow_bounds['permeability'][1])]
            self.logscale = self.logscale + [True, True]
            # activate the flow model:
            self.flow_model = FlowModel_num(self.hydraulic_df[self.Columns[:2]].values,
                                            np.geomspace(
                                                0.1, self.max_grid_ext, int(self.max_grid_ext/self.forecastDX), endpoint=True), self.layer_thick,
                                            self.flow_settings, bounds=self.FlowBounds)

        elif self.flw_model == 'num_var_perm':  # HM1 cass with FV
            self.OptimizeKeys = self.OptimizeKeys + ['zero_stress_aperture', 'residual_aperture',
                                                     'shearing_aperture_factor', 'max_shear_frac_aperture', 'shear_scaling_factor', 'storage']
            self.FlowBounds = [(self.flow_bounds['zero_stress_aperture'][0], self.flow_bounds['zero_stress_aperture'][1]),
                               (self.flow_bounds['residual_aperture'][0],
                                self.flow_bounds['residual_aperture'][1]),
                               (self.flow_bounds['shearing_aperture_factor'][0],
                                self.flow_bounds['shearing_aperture_factor'][1]),
                               (self.flow_bounds['max_shear_frac_aperture'][0],
                                self.flow_bounds['max_shear_frac_aperture'][1]),
                               (self.flow_bounds['shear_scaling_factor'][0],
                                self.flow_bounds['shear_scaling_factor'][1]),
                               (self.flow_bounds['storage'][0], self.flow_bounds['storage'][1])]
            self.logscale = self.logscale + \
                [True, True, True, True, True, True]
            # activate the flow model:
            self.flow_model = FlowModel_kchange(self.hydraulic_df[self.Columns[:2]].values,
                                                np.geomspace(0.1, self.max_grid_ext, int(self.max_grid_ext/self.forecastDX), endpoint=True),
                                                self.layer_thick,
                                                self.flow_settings,
                                                bounds=self.FlowBounds,
                                                forward_model=False,
                                                iniguess=(self.flow_settings['zero_stress_aperture'],
                                                          self.flow_settings['residual_aperture'],
                                                          self.flow_settings['shearing_aperture_factor'],
                                                          self.flow_settings['max_shear_frac_aperture'],
                                                          self.flow_settings['shear_scaling_factor'],
                                                          self.flow_settings['storage']))

        if self.seis_model == 'SEED':
            self.SEEDmodelKeys = ['seed_density', 'inter_sigma1', 'inter_sigma3', 'sigma1_std', 'sigma3_std',
                                  'cohesion_mean', 'friction_mean', 'criticality', 'stressdrop_coeff', 'phydro',
                                  'b_min', 'b_max', 'b_min_diff_stress', 'b_max_diff_stress', 'Mag_Completeness']

    def update_seeds_settings(self):
        if self.Mc == None:
            Mc = estimate_mc_maxc(self.seismicity['magnitude'].values,
                                  delta_m=self.dM,
                                  correction_factor=self.Mc_corr)
            #Mc, b_value = Calibrate_Observed_GR(
            #    self.seismicity, self.dM, 'magnitude')
            self.Mc = Mc
        else:
            Mc=self.Mc

        b_value = round(estimate_b(self.seismicity['magnitude'].values[self.seismicity['magnitude'].values >= self.Mc],
                                   self.Mc,
                                   delta_m=self.dM),
                        2)

        self.seed_settings["Mag_Completeness"] = Mc
        self.seed_settings["b_value"] = b_value
        self.b_value = b_value
        self.training_catalog = self.seismicity[
            self.seismicity['magnitude'] >= Mc]

        if (self.seed_settings["b_vs_depth_range"]):
            a, m = b_value_Lasso_calibration_1D(
                np.array([self.training_catalog['Rs'].values]),
                np.array([self.training_catalog['magnitude'].values])[:, 0],
                bins=100)
        else:
            a, m = b_value, np.zeros(1)

        self.seed_settings["a"] = a
        self.seed_settings["m"] = m

    def ll_score(self, x):
        y = self.CostFunction.y
        log_factorial_y = self.CostFunction.logyfact
        TestSettings = copy.deepcopy(self.seed_settings)

        for i in range(len(self.SEISbounds)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])

        y0 = CAPS(self.LL_Moments[1:], self.LL_Edges[self.LL_Edges >= 0.], self.Training_R,
                  self.Training_MaxPressure, TestSettings, NSim=1, retriggering=self.retriggering)[0]

        LL = np.sum(y * np.log(y0) - y0 - log_factorial_y) / \
            len(self.training_catalog)  # Poissonian LL
        if np.isfinite(LL):
            return -LL
        return 1.e20

    def ls_score(self, x):
        y = self.CostFunction.y
        TestSettings = copy.deepcopy(self.seed_settings)

        for i in range(len(self.SEISbounds)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])

        y0 = CAPS(self.LL_Moments[1:], self.LL_Edges[self.LL_Edges >= -1e-5], self.Training_R,
                  self.Training_MaxPressure, TestSettings, NSim=1, retriggering=self.retriggering)[0]

        LS = np.sum(np.square(y - y0)) / len(y)
        if np.isfinite(LS):
            if LS < self.minf:
                self.minf = LS
                self.best_x = x
            return LS
        return 1.e20

    def ls_score_seed(self, x):
        y = self.CostFunction.y
        TestSettings = copy.deepcopy(self.seed_settings)

        for i in range(len(self.SEISbounds)):
            TestSettings[self.OptimizeKeys[i]] = x[i]
            if (self.logscale[i]):
                TestSettings[self.OptimizeKeys[i]] = np.power(10., x[i])

        # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
        SeedsSettings = nb.typed.List()
        for j in range(len(self.SEEDmodelKeys)):
            SeedsSettings.append(float(TestSettings[self.SEEDmodelKeys[j]]))

        y0 = SEED(self.LL_Moments[1:], self.LL_Edges[self.LL_Edges >= -1e-5],
                  self.flow_model.r, self.flow_model.Q[:, 0],
                  self.pres_t_fit,
                  SeedsSettings, self.Nreal, self.retriggering)[1]

        y1 = np.median(y0, axis=0)

        LS = np.sum(np.square(y - y1)) / len(y)

        if np.isfinite(LS):
            if LS < self.minf:
                self.minf = LS
                self.best_x = x
            return LS
        return 1.e20

    def save_iteration_step(self, x, convergence):
        """
        function to save history of parameters and residual during DEA
        """
        # check if equal
        # np.testing.assert_equal(x, self.best_x)
        self.dea_vals.append(x)
        self.dea_func.append(self.minf)
        return

    def Calibrate_CAPS(self):
        self.update_seeds_settings()  # update b-value

        self.CostFunction = LS_cost_function_1D(np.array(self.training_catalog[['ts', 'Rs']]),
                                                bins=[self.hydraulic_df.time.values[1:],
                                                      self.forecast_edges[0]])

        # (radial) spatial bins in negative and positive directions
        self.LL_Edges = self.CostFunction.edges_x2
        self.LL_Moments = self.CostFunction.edges_x1  # time moments

        (self.Training_MaxPressure,
         self.Training_R,
         self.Training_Volume) = Interpolate_1D(self.PressureFunction, self.timesteps, self.LL_Moments, self.LL_Edges)

        self._logger.info("Maximum iterations CAPS: " + str(self.maxiter))

        # open list to save the progress of the minimization
        self.dea_vals = []
        self.dea_func = []

        x=[]

        for i in range(len(self.SEISbounds)):
            x.append(self.SEISbounds[i][0])

        result = differential_evolution(self.ls_score, self.SEISbounds, polish=True, disp=self.disp,
                                        maxiter=self.maxiter, seed=self.seed, popsize=10,
                                        updating='immediate', callback=self.save_iteration_step)
        self._logger.info('End CAPS differential evolution')

        self._logger.info('--------------------------------')
        self._logger.info('Calibrated geomechanical parameters:')
        for i in range(len(result.x)):
            self.seed_settings[self.OptimizeKeys[i]] = result.x[i]
            if (self.logscale[i]):
                self.seed_settings[self.OptimizeKeys[i]
                                   ] = np.power(10., result.x[i])
            self._logger.info(
                str(self.OptimizeKeys[i]) + ': ' + str(self.seed_settings[self.OptimizeKeys[i]]))
        self._logger.info('--------------------------------')

        self.seed_settings_optimal = dict.copy(self.seed_settings)

        self._logger.info("**************************************************")
        self._logger.info(
            "Summary of optimized parameters after CAPS calibration:")
        for i in range(len(self.SEISbounds)):
            self._logger.info(str(
                self.OptimizeKeys[i]) + ': ' + str(self.seed_settings_optimal[self.OptimizeKeys[i]]))
        j_init = len(self.SEISbounds)
        for j in range(len(self.FlowBounds)):
            self._logger.info(str(self.OptimizeKeys[j_init + j]) + ': ' + str(
                self.seed_settings_optimal[self.OptimizeKeys[j_init + j]]))
        self._logger.info("**************************************************")

        self.total_events_past_grid, self.total_rates_past_grid = CAPS(self.LL_Moments,
                                                                       self.LL_Edges[self.LL_Edges >= -1e-5],
                                                                       self.Training_R,
                                                                       self.Training_MaxPressure,
                                                                       self.seed_settings,
                                                                       NSim=self.NSim,
                                                                       retriggering=self.retriggering)

        total_events_median = np.round(
            np.median(self.total_events_past_grid, axis=0)[-1])

        self._logger.info(
            "sum of total events after calibration (median from grid): " + str(total_events_median))

        return

    def Calibrate_SEED(self):
        self.update_seeds_settings()  # update b-value

        self.CostFunction = LS_cost_function_1D(np.array(self.training_catalog[["ts", "Rs"]]),
                                                bins=[self.hydraulic_df.time.values, np.linspace(0, max(self.training_catalog['Rs'].values), 10)])

        self.LL_Edges = self.CostFunction.edges_x2
        self.LL_Moments = self.CostFunction.edges_x1

        # 1D pressure interpolation (in time only)
        pres_sol = (self.flow_model.hidden_mechanism
                    - self.hydraulic_df[self.Columns[-1]].values[0]) / (1.e6)
        self.pres_t_fit = np.array([np.interp(
            self.LL_Moments[1:], self.flow_model.Q[:, 0], ps_row) for ps_row in pres_sol.T])
        # OR:
        # Tind = np.argmin(np.abs(self.flow_model.Q[:,0] - self.LL_Moments[1:][:, np.newaxis]), axis=1)
        # self.pres_t_fit = (pres_sol[Tind,:]).T

        # Decide for the seed model version (implemented since Poster FAM23)
        if self.seed_model_version == '1':
            self.Nreal = 100  # self.NSim
        elif self.seed_model_version == '2':
            self.Nreal = 1
        else:
            self.Nreal = self.NSim

        self._logger.info("Maximum iterations SEED: " + str(self.maxiter))

        # open list to save the progress of the minimization
        self.dea_vals = []
        self.dea_func = []

        # if self.cost_fun=='LL':
        #    result = differential_evolution(self.ll_score_seed, self.SEISbounds,polish=True, disp=self.disp,
        #                                    maxiter=self.maxiter, seed=self.seed,popsize=10)
        # else:
        # if self.cost_fun=='LS':
        result = differential_evolution(self.ls_score_seed, self.SEISbounds, polish=True,
                                        disp=self.disp, maxiter=self.maxiter, seed=self.seed, popsize=10,
                                        updating='immediate', atol=1e-3, callback=self.save_iteration_step)
        self._logger.info("End SEED differential evolution")

        self._logger.info('------------------------------------')
        self._logger.info('Calibrated geomechanical parameters:')

        for i in range(len(result.x)):
            self.seed_settings[self.OptimizeKeys[i]] = result.x[i]
            if (self.logscale[i]):
                self.seed_settings[self.OptimizeKeys[i]
                                   ] = np.power(10., result.x[i])
            self._logger.info(
                str(self.OptimizeKeys[i]) + ': ' + str(self.seed_settings[self.OptimizeKeys[i]]))
        self._logger.info('------------------------------------')

        # save optimal parameters for forecast later
        self.seed_settings_optimal = dict.copy(self.seed_settings)

        self._logger.info("**************************************************")
        self._logger.info(
            "Summary of optimized parameters after CAPS calibration:")
        for i in range(len(self.SEISbounds)):
            self._logger.info(str(
                self.OptimizeKeys[i]) + ': ' + str(self.seed_settings_optimal[self.OptimizeKeys[i]]))
        j_init = len(self.SEISbounds)
        for j in range(len(self.FlowBounds)):
            self._logger.info(str(self.OptimizeKeys[j_init + j]) + ': ' + str(
                self.seed_settings_optimal[self.OptimizeKeys[j_init + j]]))
        self._logger.info("**************************************************")

        # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
        SeedsSettings = nb.typed.List()
        for j in range(len(self.SEEDmodelKeys)):
            SeedsSettings.append(
                float(self.seed_settings[self.SEEDmodelKeys[j]]))

        (self.training_catalog,
         self.total_events_past_grid,
         self.total_rates_past_grid) = SEED(self.LL_Moments[1:],
                                            self.LL_Edges[self.LL_Edges >= 0.],
                                            self.flow_model.r,
                                            self.flow_model.Q[:, 0],
                                            self.pres_t_fit,
                                            SeedsSettings,
                                            self.NSim,
                                            self.retriggering)

        total_events_median = np.round(
            np.median(self.total_events_past_grid, axis=0)[-1])

        self._logger.info("sum of total events after calibration (median from grid): "+str(total_events_median))

        return

    def run(self):
        """
        Run the model. Calls self.fit() and returns the result of
        self.predict().
        """
        self.fit()
        results = self.predict()

        results = self.parse_results(results)

        return results

    def fit(self):
        self._logger.info('Model Fitting')

        if self.flw_model == 'point_source':
            self._logger.info('Calibrating flow model (point source)....')
            self.flow_model.fit(
                self.hydraulic_df[self.Columns[::2]].values, disp=self.disp, maxiter=self.maxiter)
        elif self.flw_model == 'line_source':
            self._logger.info('Calibrating flow model (line source)....')
            self.flow_model.fit(
                self.hydraulic_df[self.Columns[::2]].values, disp=self.disp, maxiter=self.maxiter)
        elif self.flw_model == 'num_const_perm':
            self._logger.info('Calibrating flow model (numerical linear)....')
            self.flow_model.fit(self.hydraulic_df[self.Columns[:2]].values,
                                self.hydraulic_df[self.Columns[::2]].values, disp=self.disp, maxiter=self.maxiter)
        elif self.flw_model == 'num_var_perm':
            self._logger.info(
                'Calibrating flow model (numerical non-linear)....')
            self.flow_model.fit(self.hydraulic_df[self.Columns[:2]].values,
                                self.hydraulic_df[self.Columns[::2]].values, disp=self.disp, maxiter=self.maxiter)

        self._logger.info('--------------------------------')
        self._logger.info('Calibrated hydraulic parameters:')
        for i in range(len(self.flow_model.params)):
            self.seed_settings[self.OptimizeKeys[i
                                                 - len(self.flow_model.params)]] = self.flow_model.params[i]
            self._logger.info(str(
                self.OptimizeKeys[i - len(self.flow_model.params)]) + ': ' + str(self.flow_model.params[i]))
        self._logger.info('--------------------------------')

        self.timesteps = self.hydraulic_df.time.values
        inipres = self.hydraulic_df[self.Columns[-1]].values[0]

        if self.seis_model == 'CAPS':
            self._logger.info("Calibrating CAPS....")
            self.PressureFunction = create_interpolator(
                self.flow_model.r, self.timesteps, (self.flow_model.hidden_mechanism - inipres) / 1.e6)
            self.Calibrate_CAPS()
        elif self.seis_model == 'SEED':
            self._logger.info("Calibrating SEED....")
            self.Calibrate_SEED()

        Vtot = np.cumsum(
            self.hydraulic_df['topflow'].values[:-1] * np.diff(self.hydraulic_df['time'].values))
        self.af = self.b_value * self.Mc + \
            np.log10(np.median(self.total_events_past_grid, axis=0)
                     [-1]) - np.log10(Vtot)

        return

    def predict(self):
        self._logger.info('Model Predicting')
        # CAPS MODEL #

        if self.seis_model == 'CAPS':
            self.seed_settings = self.seed_settings_optimal

            (self.max_pressure_past,
             self.Forecast_R,
             self.Forecast_Volume) = Interpolate_1D(self.PressureFunction, self.timesteps, self.forecast_moments, self.forecast_edges[0])

            self.press_prediction = self.flow_model.predict(
                np.array(self.scenario_df[["time", "topflow"]].values))

            self.timesteps_predict = self.scenario_df.time.values + \
                (self.forecast_start - self.observation_start).total_seconds()
            inipres = self.hydraulic_df[self.Columns[-1]].values[0]
            self.PressureFunction_predict = create_interpolator(
                self.flow_model.r, self.timesteps_predict, (self.press_prediction - inipres) / 1.e6)

            (self.Forecast_MaxPressure,
             self.Forecast_R,
             self.Forecast_Volume) = Interpolate_1D(self.PressureFunction_predict, self.timesteps_predict, self.forecast_moments, self.forecast_edges[0])

            max_pressure = np.maximum(
                self.max_pressure_past, self.Forecast_MaxPressure)

            self._logger.info('Starting CAPS forecast')

            self.total_events_forecast_grid, self.total_rates_forecast_grid = CAPS(self.forecast_moments,
                                                                                   self.forecast_edges[0]
                                                                                   [self.forecast_edges[0] >= -1e-5],
                                                                                   self.Forecast_R,
                                                                                   max_pressure,
                                                                                   self.seed_settings,
                                                                                   NSim=self.NSim,
                                                                                   retriggering=self.retriggering)

            rate_forecast = self.total_rates_forecast_grid.T
            self._logger.info('Ending CAPS forecast')

        # SEED MODEL #
        elif self.seis_model == 'SEED':
            self.seed_settings = self.seed_settings_optimal

            self.timesteps_predict = self.scenario_df.time.values + (
                self.forecast_start - self.observation_start).total_seconds()

            self.press_prediction = self.flow_model.predict(
                np.array(self.scenario_df[["time", "topflow"]].values))

            # 1D pressure interpolation (in time only)
            pres_sol = (self.press_prediction
                        - self.hydraulic_df[self.Columns[-1]].values[0]) / (1.e6)
            self.pres_t_forecast = np.array([np.interp(self.forecast_moments, self.timesteps_predict, ps_row)
                                             for ps_row in pres_sol.T])

            # OR:
            # Tind = np.argmin(np.abs(PlannedStrategy['time'].values - Forecasts_Moments[:, np.newaxis]), axis=1)
            # self.pres_t_fit = (pres_sol[Tind,:]).T

            # typed.List instance which is a compiled library backed, type-homogeneous list data type that is an improvement over the reflected-list: SeedsSettings = []
            SeedsSettings = nb.typed.List()
            for j in range(len(self.SEEDmodelKeys)):
                SeedsSettings.append(
                    float(self.seed_settings[self.SEEDmodelKeys[j]]))

            (self.synthetic_catalog,
             self.total_events_grid,
             self.total_rates_grid) = SEED(np.concatenate((self.LL_Moments[1:], self.forecast_moments[1:]), axis=0),
                                           self.forecast_edges[0][self.forecast_edges[0] >= 0.],
                                           self.flow_model.r,
                                           np.concatenate(
                                               (self.flow_model.Q[:, 0], self.scenario_df['time'].values[1:]), axis=0),
                                           np.concatenate(
                                               (self.pres_t_fit, self.pres_t_forecast[:, 1:]), axis=1),
                                           SeedsSettings,
                                           self.NSim,
                                           self.retriggering)

            self.total_events_past_grid = self.total_events_grid[:, :len(
                self.LL_Moments[1:])]
            self.total_events_forecast_grid = self.total_events_grid[:, (len(
                self.LL_Moments[1:]) - 1):]

            rate_forecast = np.diff(self.total_events_forecast_grid, axis=1).T

        events_forecast_grid = (self.total_events_forecast_grid).T
        b_value = self.b_value
        Mc = self.Mc
        a_value = np.log10(events_forecast_grid) + b_value * Mc
        volume_past = np.cumsum(
            self.hydraulic_df['topflow'].values[:-1] * np.diff(self.hydraulic_df['time'].values))
        volume = volume_past[-1] + np.cumsum(
            self.scenario_df['topflow'].values[:-1] * np.diff(self.scenario_df['time'].values))
        alpha = np.log10(events_forecast_grid[-1] / volume[-1]) + b_value * Mc

        self.dt_fit = np.array(
            (self.hydraulic_df.index.values
             - self.hydraulic_df.index.values[0]).tolist())[1:] / 1.e9

        NN = ((np.max(self.dt_fit) - np.min(self.dt_fit))
              / self.parameters['epoch_duration']).astype(int) + 1
        self._time_fit = np.linspace(
            np.min(self.dt_fit), np.max(self.dt_fit), NN)

        rate_fit = np.diff(np.concatenate(
            (np.array([0,]),
             np.interp(self._time_fit, self.dt_fit, np.median(self.total_events_past_grid, axis=0)))
        ))

        fit_datetimes = [self.observation_start + timedelta(seconds=t)
                         for t in self._time_fit]

        return a_value, alpha, b_value, rate_fit, rate_forecast

    def parse_results(self, results):
        """
        Parse Results to a ForecastGRRateGrid per timestep.
        A ForecastGRRateGrid requires at least the following columns:
            'number_events', 'longitude_min', 'longitude_max',
            'latitude_min', 'latitude_max', 'depth_min', 'depth_max',
            'a', 'b', 'mc', 'grid_id'

        """
        fit_datetimes = [self.observation_start + timedelta(seconds=t)
                         for t in self._time_fit]
        forecast_datetimes = [self.observation_start + timedelta(seconds=t)
                              for t in self.forecast_moments]

        bounds = list(self.bounding_polygon.bounds)
        bounds.insert(5, self.depth_min)
        bounds.insert(2, self.depth_max)

        parsed_results = []

        for i in range(len(fit_datetimes) - 1):
            rate = {'number_events': results[3][i+1],
                    'longitude_min': bounds[0],
                    'longitude_max': bounds[3],
                    'latitude_min': bounds[1],
                    'latitude_max': bounds[4],
                    'depth_min': bounds[5],
                    'depth_max': bounds[2],
                    # self.af,
                    'a': np.log10(len(self.seismicity[self.seismicity['magnitude'] > self.Mc])) + self.b_value * self.Mc,
                    'alpha': self.af,
                    'b': self.b_value,
                    'mc': self.Mc,
                    'grid_id': 0}

            # Create a ForecastGRRateGrid for each Timestep
            rategrid = ForecastGRRateGrid(rate,
                                          starttime=fit_datetimes[i],
                                          endtime=fit_datetimes[i + 1])

            parsed_results.append(rategrid)

        for i in range(len(forecast_datetimes) - 1):
            rate = {'number_events': results[4][i],
                    'longitude_min': [bounds[0]] * self.NSim,
                    'longitude_max': [bounds[3]] * self.NSim,
                    'latitude_min': [bounds[1]] * self.NSim,
                    'latitude_max': [bounds[4]] * self.NSim,
                    'depth_min': [bounds[5]] * self.NSim,
                    'depth_max': [bounds[2]] * self.NSim,
                    'a': results[0][i],
                    'alpha': results[1],
                    'b': results[2],
                    'mc': [self.Mc] * self.NSim,
                    'grid_id': list(range(self.NSim))}

            # Create a ForecastGRRateGrid for each Timestep
            rategrid = ForecastGRRateGrid(rate,
                                          starttime=forecast_datetimes[i],
                                          endtime=forecast_datetimes[i + 1])

            parsed_results.append(rategrid)

        return parsed_results
