# Copyright 2018, ETH Zurich - Swiss Seismological Service SED
"""
Miscellaneous HM1 model core facilities.
"""

import matplotlib.cm as cm
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.colors import Normalize
from scipy import special, stats
from scipy.interpolate import CloughTocher2DInterpolator
from sklearn import linear_model


def create_interpolator(flow_model_r, t_grid, solution):
    if len(flow_model_r) == len(solution[0]):
        mesh = np.meshgrid(flow_model_r, t_grid)
    else:
        mesh = np.meshgrid(flow_model_r[1:], t_grid)
    interpd = CloughTocher2DInterpolator(
        np.array([mesh[0].ravel(), mesh[1].ravel()]).T, solution.ravel(), fill_value=0)
    return interpd


def Interpolate_1D(PressureFunction, timesteps, Forecasts_Moments, Forecasts_Edges):
    """
    Same as CAPS.InterpolateMaximumPressure, but using a self.PressureFunction created by the fit
    """
    RR = 1  # CAPS_KD_Settings["RefinementRatio"]
    # for 1D we use directly the cartesian grid as the radius
    MyR = Refine_An_Edge(Forecasts_Edges, RR)
    LaterKeepIndices = np.logical_and(np.isfinite(MyR), MyR > 0.)
    My_Working_R = MyR[LaterKeepIndices]
    DT = np.concatenate((timesteps, Forecasts_Moments))
    MyT = np.sort(np.unique(DT))
    MyPressure, MaxPressureAll = np.zeros([2, len(MyT), len(My_Working_R)])
    PM = np.zeros(len(My_Working_R))
    for timeIter in range(len(MyT)):
        ConsiderTime = MyT[timeIter]
        MyPressure[timeIter, :] = PressureFunction(My_Working_R, ConsiderTime)
        MaxPressureAll[timeIter, :] = np.maximum(PM, MyPressure[timeIter, :])
        PM = MaxPressureAll[timeIter, :]

    TimeBinIndeces = np.intersect1d(Forecasts_Moments, MyT, return_indices=True)[
        2]  # discard unecessary time discretization now
    MaxPressure = MaxPressureAll[TimeBinIndeces]
    dV = np.unique(np.diff(Forecasts_Edges, axis=0))
    # dV=np.unique(np.diff(Forecasts_Edges, axis=0), axis=0)
    # if(len(dV)>1):
    # print ("Warning: non equidistant edges have been passed in the one direction. Correct array of control distances.")
    # \todo: assume equidistant edges and refinement
    My_Cell_Width = np.prod(np.amin(dV, axis=0) / RR)
    return MaxPressure, My_Working_R, np.ones(len(My_Working_R)) * My_Cell_Width


def daterange(start_date,
              end_date,
              epoch_duration=None,
              training_start_date=None):
    """Numpy array returned with forecasting intervals in seconds
    increasing with constant increment, inclusive of the end_date.

    :param st art_date: Datetime for generator to start with.
    :param end_date: Datetime for generator to finish with.
    :param epoch_duration: Number of seconds between returned values.

    :rtype: np.array of seconds from training start time to forecast for
    """
    total_seconds = (end_date - start_date).total_seconds()
    if not epoch_duration:
        epoch_duration = total_seconds

    # the +1 is required to match the r code which gives the range
    # including the final variable.
    daterange_list = [
        i * epoch_duration for i
        in range(int(total_seconds / epoch_duration) + 1)]
    if training_start_date:
        daterange_list = [i + (
            start_date - training_start_date).total_seconds()
            for i in daterange_list]
    date_range_array = np.array(daterange_list)
    return date_range_array


def Calibrate_Observed_GR(catalog, dM=0.02, col_name='Mw'):
    """
     Very similar to ramsis.EM1 mc_mode
     The Frequency Mass Density (FMD) is found by smoothing
     observed Magnitudes (M) with Gaussian kernels. Then
     the completion Magnitude (Mc) is set to the rounded value
     at the peak of the FMD.

     The slope of the Gutenberg-Richter law is returned with
     the MLE method, Aki (1965) and for the found Mc.

     NOTE: It returns Nan when less than 4 events are passed.
      The scientific implications behind it need to be further
      studied (the same applies for the ramsis.EM1). E.g. binomial
      is more possible; here the mode with the largest M is chosen.

     :param Magnitudes: magnitudes of incomplete catalog
     :param dM        : size of binning and of rounding Mc

     :rtype: float for the Mc
     :rtype: float for the b-value
     """
    # print("starting the GR")
    Magnitudes = np.array(catalog[col_name].values)
    Mc, b = np.full(2, np.nan)
    # print("inputs to gr: ", catalog)
    # print("len magnitudes: ", len(Magnitudes))
    if (len(Magnitudes) > 4):
        Magnitudes = Magnitudes[~np.isnan(Magnitudes)]
        Mrange = np.around(np.arange(np.amax(Magnitudes) + 5 * dM, np.amin(Magnitudes) - 5 * dM, -dM),
                           decimals=int(-np.log(dM)))
        argMc = np.argmax(stats.gaussian_kde(Magnitudes)(Mrange))
        # Mc = round(Mrange[argMc]/dM)*dM
        Mc = Mrange[argMc]
        # print("mc", Mc)
        #        b=np.log10(np.exp(1.))/(np.mean(Magnitudes[np.where(Magnitudes>=Mc)])-np.min(Magnitudes[np.where(Magnitudes>=Mc)]))
        b = np.log10(
            np.exp(1.)) / (np.mean(Magnitudes[np.where(Magnitudes >= Mc)]) - (Mc - dM / 2))
        # print('b', b)
    return Mc, b


def b_value_Lasso_calibration_1D(Xs, Mw, bins=100):
    Xs = Xs[:, 0]  # this enables the right dimensions for the histogram (1-D)
    counts, edges = np.histogram(Xs, bins=bins)
    counts = counts.flatten()  # returns a copy collapsed into one dimension
    meanMw, edges = np.histogram(Xs, bins=edges, weights=Mw)
    meanMw = meanMw.flatten()

    BinsGrid = np.meshgrid(edges[:-1] + 0.5 * (edges[1:] - edges[:-1]))

    X = np.array([BinsGrid[0].flatten()])

    Keep = np.where(counts > 0)
    X = X[:, Keep[0]].transpose()
    Y = meanMw[Keep] * np.reciprocal(counts[Keep])

    clf = linear_model.Lasso(alpha=0.1)
    clf.fit(X, Y)
    return clf.intercept_, clf.coef_


def Refine_An_Edge(Coarse_Edges_1D, RR=5):
    Coarse_Cells = np.sort(
        np.unique(Coarse_Edges_1D.ravel()), kind='mergesort')
    MyFinerCells = np.zeros((len(Coarse_Cells) - 1) * RR)
    for i in range(len(Coarse_Cells) - 1):
        Delta = (Coarse_Cells[i + 1] - Coarse_Cells[i]) / RR
        for ii in range(RR):
            MyFinerCells[RR * i + ii] = Coarse_Cells[i] + Delta * (ii + 0.5)
    return MyFinerCells

class LL_cost_function_1D:
    def __init__(self, x, bins=(10, 10, 10, 10)):
        self.y, self.edges_x1, self.edges_x2 = np.histogram2d(
            x[:, 0], x[:, 1], bins=bins, density=False)
        self.y = np.sum(self.y, axis=1)
        # self.y = np.sum(self.y,axis=1)
        self.logyfact = np.log(special.factorial(self.y)).ravel()

    def score(self, y1):
        ll = self.y * np.log(y1.ravel()) - y1.ravel() - self.logyfact
        return np.sum(ll)


class LS_cost_function_1D:
    def __init__(self, x, bins=(10, 10)):
        self.y, self.edges_x1, self.edges_x2 = np.histogram2d(
            x[:, 0], x[:, 1], bins=bins, density=False)
        self.y = np.cumsum(np.sum(self.y, axis=1))

    def score(self, y1):
        ls = np.sum(np.square(self.y - y1))
        return np.nanmean(ls)


def forecast_space_bins(space, DX):
    """
    Same as above but only for 1D Bin forecasting and only considering one dimension,
    either X, Y, or Z.
    """

    Nx = int((space[1] - space[0]) / DX)
    if (Nx % 2) == 0:
        Nx = Nx
    else:
        Nx = Nx + 1
    ForSpace_bins = np.linspace(space[0], space[1], num=Nx + 1)

#  Necessary Input for Forecast
    Forecasts_Edges = [ForSpace_bins]
    return Forecasts_Edges


def plotting_data(model):
    fig, ax = plt.subplots(1, 2, figsize=(20, 6))

    ax[0].scatter(model.seismicity.time, (model.seismicity['magnitude'] - np.min(model.seismicity['magnitude']))
                  / (np.max(model.seismicity['magnitude'] - np.min(model.seismicity['magnitude']))) / 4,
                  color='lightgrey', alpha=.5, zorder=2, label='seismics')
    ax[0].step(model.hydraulic_df.index, model.hydraulic_df['topflow']
               / np.max(model.hydraulic_df['topflow']) / 4, color='darkblue', label='hydraulics[topflow]')
    ax[0].step(model.hydraulic_df.index, model.hydraulic_df['toppressure']
               / np.max(model.hydraulic_df['toppressure']) / 4, color='darkred', label='hydraulics[toppressure]')
    ax[0].set_xlabel('Date')
    ax[0].set_ylabel('Normalised inputs')
    ax[0].set_title('Training data')
    ax[0].legend(loc='upper left')
    ax[0].spines['right'].set_visible(False)
    ax[0].spines['top'].set_visible(False)

    ax[1].plot(model.hydraulic_df.index, model.hydraulic_df['topflow'], color='darkblue',
               label='hydraulics[topflow]')
    ax[1].plot(model.scenario_df.index, model.scenario_df[model.column_flow],
               color='royalblue', label='scenario')
    ax[1].set_xlabel('Date')
    ax[1].set_ylabel('Flow rate [m$^3$/s]')
    ax[1].set_title('Flow time series')
    ax[1].legend()
    ax[1].spines['right'].set_visible(False)
    ax[1].spines['top'].set_visible(False)
    fig.tight_layout()


def plotting_Pfit(model):
    fig, ax = plt.subplots(1, 2, figsize=(20, 6))

    ax[0].step(model.hydraulic_df.index, model.hydraulic_df['toppressure']
               / 1.e6, color='darkred', label='hydraulics[toppressure]')
    ax[0].plot(model.hydraulic_df.index, model.flow_model.hidden_mechanism[:,
               0] / 1.e6, color='darkblue', label='modeled pressure')

    ax[0].set_xlabel('Date')
    ax[0].set_ylabel('Pressure [MPa]')
    ax[0].set_title('Fitting pressure')
    ax[0].legend(loc='upper left')
    ax[0].spines['right'].set_visible(False)
    ax[0].spines['top'].set_visible(False)

    time_vector = np.linspace(model.hydraulic_df.time.values[0],
                              model.hydraulic_df.time.values[-1],
                              len(model.hydraulic_df))  # Example time vector

    # Normalize the vector to be between 0 and 1 for colormap mapping
    norm = Normalize(vmin=np.min(time_vector), vmax=np.max(time_vector))

    # Choose a colormap (e.g., 'viridis', 'plasma', 'coolwarm')
    cmap = cm.viridis

    # Loop over the matrix rows and plot each line with a color from the colormap
    for i in range(len(model.hydraulic_df)):
        # Get color corresponding to the time vector
        color = cmap(norm(time_vector[i]))
        ax[1].plot(model.flow_model.r,
                   model.flow_model.hidden_mechanism[i, :] / 1.e6, color=color)

    ax[1].set_xlabel('Distance [m]')
    ax[1].set_ylabel('Pressure [MPa]')
    ax[1].set_title('Space/time distribution')
    ax[0].spines['right'].set_visible(False)
    ax[0].spines['top'].set_visible(False)

    fig.tight_layout()


def plotting_results(model, model_output, seism_df=pd.DataFrame()):

    inj_t = np.array((model.hydraulic_df.index.values
                     - model.hydraulic_df.index.values[0]).tolist()) / 1.e9
    Dtau = model.model_input.model_settings['epoch_duration']

    NN = ((inj_t[-1] - inj_t[0])
          / Dtau).astype(int) + 1
    time_fit = np.linspace(
        inj_t[0], inj_t[-1], NN)

    if len(model_output[3]) < len(time_fit):
        time_fit_mids = time_fit[1:]
    else:
        time_fit_mids = time_fit

    time_forecast_mids = model.forecast_moments[1:] - Dtau / 2

    seism = model.seismicity
    seism = seism[seism['magnitude'] >= model.Mc]
    seism = seism.copy()
    seism.sort_values(by='time', inplace=True)
    seism_t = np.array(
        (seism['time'].values - model.hydraulic_df.index.values[0]).tolist()) / 1.e9

    # range pred

    y_pred_mean = np.median(model_output[4], axis=1)
    up = np.percentile(model_output[4], axis=1, q=75)
    down = np.percentile(model_output[4], axis=1, q=25)
    down[down < 0] = 0

    if not seism_df.empty:
        seism_t_all = np.array(
            (seism_df['time'].values - model.hydraulic_df.index.values[0]).tolist()) / 1.e9
        ind = np.logical_and(seism_t_all > 0, seism_t_all
                             < model.forecast_moments[-1])
        seism_df_filt = seism_df[ind]
        seism_t_all = seism_t_all[ind]
        seism_t_all_aboveMc = seism_t_all[seism_df_filt['magnitude'] >= model.Mc]
        taubins_all = np.arange(0, model.forecast_moments[-1] + Dtau, Dtau)
        tmids_all = taubins_all[1:] - Dtau / 2
        rate_all = np.histogram(seism_t_all_aboveMc, bins=taubins_all)

    ##
    fig, ax = plt.subplots(1, 2, figsize=(20, 6))

    ax[0].scatter(seism_t, np.arange(1, len(seism_t) + 1),
                  color='darkgrey', label='observations')
    ax[0].plot(time_fit_mids, np.cumsum(model_output[3]),
               color='darkgreen', label='model fit')
    ax[0].set_xlabel('Time increment [s]')
    ax[0].set_ylabel('Cumulative number of events')
    ax[0].set_title('Training fit')
    ax[0].legend()
    ax[0].spines['right'].set_visible(False)
    ax[0].spines['top'].set_visible(False)

    # ax[1].plot(model.res_fit_t, model.rate_train, color = 'lightgrey', label = 'observations')
    ax[1].plot(time_fit_mids, model_output[3],
               color='darkgreen', label='model fit')
    ax[1].plot(time_forecast_mids, y_pred_mean,
               color='darkred', label='model prediction')
    ax[1].scatter(time_forecast_mids, y_pred_mean, color='darkred')
    ax[1].fill_between(time_forecast_mids, down, up,
                       facecolor='darkred', alpha=.3)
    if not seism_df.empty:
        ax[1].step(tmids_all, rate_all[0], where='mid',
                   label='data', color='darkgrey')
    # ax[1].fill_between(tmids_all, 0, rate_all[0], color='grey', alpha = .5)
    # ax[1].set_xlim(0, 50000)
    # ax[1].set_ylim(0, 150)
    ax[1].set_xlabel('Time increment [s]')
    ax[1].set_ylabel('Event rate [time bin]')
    ax[1].set_title('Event-rate time series')
    ax[1].legend(loc='upper left')
    ax[1].spines['right'].set_visible(False)
    ax[1].spines['top'].set_visible(False)
    # fig.tight_layout()
