import logging
import os
from datetime import datetime

import numpy as np
import pandas as pd
from hydws.parser import BoreholeHydraulics
from seismostats import Catalog
from seismostats.utils.coordinates import CoordinateTransformer
from shapely import Polygon, to_wkt
from shapely.wkt import dumps

from hm1d.hermes import ModelInput, run_hm1d, run_hm1d_validated  # noqa
from hm1d.model import HM1D
from hm1d.utils import plotting_data, plotting_Pfit, plotting_results

import matplotlib
matplotlib.use('TkAgg')

# Set up logger
date_strftime_format = "%d-%m-%y %H:%M:%S"
logging.basicConfig(
    format='%(asctime)s %(levelname)s  %(message)s',
    datefmt=date_strftime_format,
    level=logging.INFO)


# def main():
############################################################
# HYDRAULICS
dirpath = os.path.dirname(os.path.abspath(__file__))
Data_Path = os.path.join(dirpath, "data/")
# hydraulics = pd.read_hdf(Data_Path + 'TrainingData.h5')
hydraulics = pd.read_hdf(Data_Path + 'Hydr_FORGE22.h5')
hydraulics_plan = hydraulics[hydraulics.index.values >= np.array(
    datetime(2022, 4, 21, 16,29)).astype('datetime64')]
hydraulics_plan = hydraulics_plan[hydraulics_plan.index.values < np.array(
    datetime(2022, 4, 21, 17,00)).astype('datetime64')]


# define start and end of forecast
# datetime.utcfromtimestamp(hydraulics_plan.index.values[0].tolist() / 1e9)
forecast_start = datetime(2022, 4, 21, 16,20)
# datetime.utcfromtimestamp(hydraulics_plan.index.values[-1].tolist() / 1e9)
forecast_end = datetime(2022, 4, 21, 17, 00)

hydraulics = hydraulics[hydraulics.index.values >= np.array(
    datetime(2022, 4, 21, 13, 00, 23)).astype('datetime64')]
hydraulics = hydraulics[hydraulics.index.values < np.array(
    datetime(2022, 4, 21, 16,30)).astype('datetime64')]

#hydraulics_plan["topflow"]=-1*np.ones((len(hydraulics_plan),1))
hydraulics_plan.index.values[-1] = np.array(
    datetime(2022, 4, 21, 17,00)).astype('datetime64')
hydraulics_plan.topflow.values[-1] = hydraulics_plan.topflow.values[-2]

hydraulics_plan = hydraulics_plan[~hydraulics_plan.index.duplicated(
    keep='first')]

# hydraulics_plan = pd.read_hdf(Data_Path + 'Scenario.h5')


observation_start = datetime.utcfromtimestamp(
    hydraulics.index.values[0].tolist() / 1e9)
observation_end = datetime.utcfromtimestamp(
    hydraulics.index.values[-1].tolist() / 1e9)

# rename columns to match hydws
# hydraulics = hydraulics.drop('time', axis=1).rename(
#    columns={'pressure': 'toppressure'})
hydraulics_plan.index.names = ['datetime']

# generate empty parser and load dataframe to empty section
parser = BoreholeHydraulics()
section_id = parser.section_from_dataframe(hydraulics)
hydjson = parser.to_json()

# replace hydraulics with scenario hydraulics
parser[section_id].hydraulics = hydraulics_plan
scenario_hydjson = parser.to_json()


############################################################
# SEISMICITY

proj_string = 'epsg:32143'
refGES = [
    -112.8963897,
    38.50402147,
    -1650.0249
]
# define coordinate transformer, local <-> swiss, external <-> WGS84
transformer = CoordinateTransformer(proj_string)
top_well = transformer.to_local_coords(*refGES)

# CoordinateTransformer with reference point in epsg:32143
transformer = CoordinateTransformer(proj_string, *top_well)

# Read seismics, transform to WGS84
# catalog_df = pd.read_hdf(Data_Path + 'TrainingTarget.h5')
catalog_df = pd.read_hdf(Data_Path + 'Cat_FORGE22.h5')
catalog_df = catalog_df[catalog_df.index.values
                        < np.array(observation_end).astype('datetime64')]
catalog_df = catalog_df[catalog_df.index.values
                        >= np.array(observation_start).astype('datetime64')]

#convert from local coordinates with respect to well pad to lat/lon (although the file should have it already)
catalog_df['Xs'], catalog_df['Ys'], catalog_df['Zs'] = \
    transformer.from_local_coords(catalog_df.X.values,
                                  catalog_df.Y.values,
                                  catalog_df.Z.values)

catalog_df.index.names = ['datetime']
catalog_df = catalog_df.reset_index().rename(
    columns={
        'Xs': 'longitude',
        'Ys': 'latitude',
        'Zs': 'depth',
        'mag': 'magnitude',
        'datetime': 'time'})

catalog = Catalog(catalog_df)
catalog['magnitude_type'] = 'mag'

quakeml = catalog.to_quakeml()

reference = [-112.88512, 38.50183, 846.28710] # location check shot 

bounding_polygon = Polygon([[-150, -150], [-150, 150], [150, 150],
                           [150, -150], [-150, -150]])
bounding_polygon = transformer.polygon_from_local_coords(bounding_polygon)

data = {
    'forecast_start': forecast_start,
    'forecast_end': forecast_end,
    'injection_observation': [hydjson],
    'injection_plan': [scenario_hydjson],
    'bounding_polygon': to_wkt(bounding_polygon),
    'depth_min': 0,
    'depth_max': 10,
    'seismicity_observation': quakeml,
    'model_parameters': {
        'Delta_tau': 30,
        'forecast_dx':20,
        'max_iter_DE': 10,
        'DE_verbose': True,
        'SeisSettings': {
            'seis_model': 'CAPS',  # SEED or CAPS
            'NSim': 100,  # SEED & CAPS
            'seed_model_version': 2,  # SEED
            "seed_density": 1,  # SEED & CAPS
            "inter_sigma1": 63.9,  # SEED & CAPS
            "inter_sigma3": 41.85,  # SEED & CAPS
            "sigma1_std": 6.4,  # SEED & CAPS
            "sigma3_std": 4.2,  # SEED & CAPS
            "cohesion_mean": 3,  # SEED & CAPS
            "friction_mean": 0.5,  # SEED & CAPS
            "criticality": 0.0,  # SEED
            "stressdrop_coeff": 0.1,  # SEED & CAPS
            "phydro": 24.28,  # SEED & CAPS
#            "b_min": 1.0,  # SEED
#            "b_max": 5.0,  # SEED
#            "b_min_diff_stress": 50,  # SEED
#            "b_max_diff_stress": 0,  # SEED
#            "b_vs_depth_range": True,  # SEED
            "CAPS_sigma_distr": 0.05,  # CAPS
            "retriggering": True,
            'dM': 0.05,
            'Mc':None,
            'mc_corr':0.2,
            'SeisBounds': {
                "sigma1_std": (0.5, 7.5),
                "sigma3_std": (0.5, 5.5),
                "stressdrop_coeff": (1.e-7, 1.e-2),
                "cohesion_mean": (3.0, 3.0),
                "friction_mean": (0.5, 0.75),
                "seed_density": (3000, 3500),
                #"sigma1_std": [4.400356761344377, 4.400356761344377],
                #"sigma3_std": [2.124334139834986, 2.124334139834986],
                #"stressdrop_coeff": [0.0069878723671454, 0.0069878723671454],
                #"cohesion_mean": [3, 3],
                #"friction_mean": [0.5458226901169759, 0.5458226901169759],
                #"seed_density": [3473.996592550586, 3473.996592550586]
            },
        },
        'FlowSettings': {
            'flow_model': 'num_var_perm',
            # 'point_source' , 'line_source', 'num_const_perm', 'num_var_perm_FD', 'num_var_perm'
            "permeability": -20,  # 'point_source' , 'line_source', 'num_const_perm'
            # 'point_source' , 'line_source', 'num_const_perm', 'num_var_perm_FD', 'num_var_perm'
            "storage": -8.5,
            "viscosity": 1e-3,
            # 'point_source' , 'line_source', 'num_const_perm', 'num_var_perm_FD', 'num_var_perm'
            'zero_stress_aperture': -4.1,  # 'num_var_perm'
            'residual_aperture': -6.9,  # 'num_var_perm'
            'frac_normal_stiffness': 12,  # 'num_var_perm'
            'shearing_aperture_factor': 0,  # 'num_var_perm'
            'max_shear_frac_aperture': 0,  # 'num_var_perm'
            'shear_scaling_factor': 0,  # 'num_var_perm'
            'damping_parameter': 0.6,  # 'num_var_perm'
            'threshold_pressure': 45.18,  # 'num_var_perm'
            'fix_normal_stress': 50.65,  # 'num_var_perm'
            'frac_spacing': 1,  # 'num_var_perm'
            # default to 300, 'point_source' , 'line_source', 'num_const_perm',, 'num_var_perm'
            'max_grid_ext': 1000,
            'layer_thick': 6,  # default to 100, 'line_source', 'num_const_perm', 'num_var_perm'
            'cross_area': 94,  # default to 94, 'point_source
            'FlowBounds': {
                # 'point_source' , 'line_source', 'num_const_perm', 'num_var_perm'
                "storage": (-8.3, -8.2),
                'zero_stress_aperture': (-5, -2),  # 'num_var_perm'
                'residual_aperture': (-8, -5),  # 'num_var_perm'
                'shearing_aperture_factor': (-9, -4),  # 'num_var_perm'
                'max_shear_frac_aperture': (-7, -4),  # 'num_var_perm'
                'shear_scaling_factor': (-13, -8),  # 'num_var_perm'
                #"storage": [-8.299934237884381, -8.299934237884381],
                #"zero_stress_aperture": [-3.97449816183616, -3.97449816183616],
                #"residual_aperture": [-5.141763467975515, -5.141763467975515],
                #"shearing_aperture_factor": [-5.107858970102068, -5.107858970102068],
                #"max_shear_frac_aperture": [-4.214434994790424, -4.214434994790424],
                #"shear_scaling_factor": [-12.166857388340654, -12.166857388340654]
            },
        }
    },
    "model_settings": {
        "well_section_id": str(section_id),
        "epoch_duration": 300,
        "n_phases": 12,
        "injection_point": reference,  # always in WGS84
        "local_proj_string": "epsg:32143"
    },
}

model_input = ModelInput(**data)

model = HM1D(model_input)

plotting_data(model)

model.fit()

plotting_Pfit(model)

results = model.predict()


# results = model.parse_results(results)
# catalog_df_full = pd.read_hdf(Data_Path + 'seismic_Data_full_located.h5')
catalog_df_full = pd.read_hdf(Data_Path + 'Cat_FORGE22.h5')
catalog_df_full.index.names = ['datetime']
catalog_df_full = catalog_df_full.reset_index().rename(
    columns={
        'long': 'longitude',
        'lat': 'latitude',
        'altitude': 'depth',
        'mag': 'magnitude',
        'datetime': 'time'})

# catalog_df_full.rename(columns={"X": "Xs", "Y": "Ys", "Depth": "Zs", "MomMag": "magnitude", "Time(UTC)": "time"}, inplace=True)
# catalog_df_full.rename(columns={"mag": "magnitude"}, inplace=True)
catalog_df_full = catalog_df_full[catalog_df_full["magnitude"] > -5]
catalog_df_full.set_index("time", inplace=True)
catalog_df_full = catalog_df_full[catalog_df_full.index
                                  >= hydraulics.index.values[0]]
catalog_df_full = catalog_df_full[catalog_df_full.index <= forecast_end]
catalog_df_full = catalog_df_full.reset_index()

plotting_results(model, results, catalog_df_full)

# return model


# if __name__ == "__main__":
#    result = main()


#hydraulics = pd.read_hdf(Data_Path + 'test.h5')
#hydraulics.rename(columns={"flow": "topflow", "pressure": "toppressure"}, inplace=True)
#hydraulics.drop(columns={"time"}, inplace=True)
#import json
#with open('result.json', 'w') as fp:
#    json.dump(hydjson, fp, indent=4)
