import logging
import os
from datetime import datetime

import pandas as pd
from hydws.parser import BoreholeHydraulics
from seismostats import Catalog
from seismostats.utils.coordinates import CoordinateTransformer
from shapely import Polygon, to_wkt

from hm1d.hermes import run_hm1d, run_hm1d_validated  # noqa

# Set up logger
date_strftime_format = "%d-%m-%y %H:%M:%S"
logging.basicConfig(
    format='%(asctime)s %(levelname)s  %(message)s',
    datefmt=date_strftime_format,
    level=logging.INFO)


def main():
    ############################################################
    # HYDRAULICS
    dirpath = os.path.dirname(os.path.abspath(__file__))
    Data_Path = os.path.join(dirpath, "data/")
    hydraulics = pd.read_hdf(Data_Path + 'TrainingData.h5')
    hydraulics = hydraulics.resample('600s').mean().interpolate()

    hydraulics_plan = pd.read_hdf(Data_Path + 'Scenario.h5')

    # define start and end of forecast
    forecast_start = datetime.utcfromtimestamp(
        hydraulics_plan.index.values[0].tolist() / 1e9)
    forecast_end = datetime.utcfromtimestamp(
        hydraulics_plan.index.values[-1].tolist() / 1e9)

    # rename columns to match hydws
    hydraulics = hydraulics.drop('time', axis=1).rename(
        columns={'pressure': 'toppressure'})
    hydraulics_plan.index.names = ['datetime']

    # generate empty parser and load dataframe to empty section
    parser = BoreholeHydraulics()
    section_id = parser.section_from_dataframe(hydraulics)
    hydjson = parser.to_json()

    # replace hydraulics with scenario hydraulics
    parser[section_id].hydraulics = hydraulics_plan
    scenario_hydjson = parser.to_json()

    ############################################################
    # SEISMICITY

    proj_string = 'epsg:2056'

    # define coordinate transformer, local <-> swiss, external <-> WGS84
    reference = [8.47449792444771,  # longitude
                 46.5098187019071,  # latitude
                 1271.43402303251]

    transformer = CoordinateTransformer(proj_string)
    top_section_swiss = transformer.to_local_coords(*reference)

    # CoordinateTransformer with reference point in epsg:2056
    transformer = CoordinateTransformer(proj_string, *top_section_swiss)

    # Read seismics, transform to WGS84
    catalog_df = pd.read_hdf(Data_Path + 'TrainingTarget.h5')

    catalog_df['Xs'], catalog_df['Ys'], catalog_df['Zs'] = \
        transformer.from_local_coords(catalog_df.Xs.values,
                                      catalog_df.Ys.values,
                                      catalog_df.Zs.values)

    catalog_df = catalog_df.reset_index().rename(
        columns={
            'Xs': 'longitude',
            'Ys': 'latitude',
            'Zs': 'depth',
            'mag': 'magnitude',
            'datetime': 'time'})

    catalog = Catalog(catalog_df)
    catalog['magnitude_type'] = 'Mw'

    quakeml = catalog.to_quakeml()

    bounding_polygon = Polygon([[-150, -150], [-150, 150], [150, 150],
                               [150, -150], [-150, -150]])
    bounding_polygon = transformer.polygon_from_local_coords(bounding_polygon)

    data = {
        'forecast_start': forecast_start,
        'forecast_end': forecast_end,
        'injection_observation': [hydjson],
        'injection_plan': [scenario_hydjson],
        'bounding_polygon': to_wkt(bounding_polygon),
        'depth_min': 150 + top_section_swiss[2],
        'depth_max': -150 + top_section_swiss[2],
        'seismicity_observation': quakeml,
        'model_parameters': {
            'Delta_tau': 600,  # resample DT hydraulic data training
            'max_iter_DE': 100,
            'DE_verbose': True,
            'SeisSettings': {
                'seis_model': 'CAPS',  # SEED or CAPS
                'NSim': 50,  # SEED & CAPS
                'seed_model_version': 2,  # SEED
                "seed_density": 5,  # SEED & CAPS
                "inter_sigma1": 26.5,  # SEED & CAPS
                "inter_sigma3": 16.7,  # SEED & CAPS
                "sigma1_std": 20,  # SEED & CAPS
                "sigma3_std": 12,  # SEED & CAPS
                "cohesion_mean": 1.0,  # SEED & CAPS
                "friction_mean": 0.6,  # SEED & CAPS
                "criticality": 0,  # SEED
                "stressdrop_coeff": 0.1,  # SEED & CAPS
                "phydro": 4.0,  # SEED & CAPS
                "b_min": 1.0,  # SEED
                "b_max": 5.0,  # SEED
                "b_min_diff_stress": 50,  # SEED
                "b_max_diff_stress": 0,  # SEED
                "b_vs_depth_range": True,  # SEED
                "CAPS_sigma_distr": 0.05,  # CAPS
                "retriggering": True,
                'dM': 0.05,
                'SeisBounds': {
                    "sigma1_std": (0.01, 10),
                    "sigma3_std": (0.01, 10),
                    "stressdrop_coeff": (1.e-3, 0.01),
                    "cohesion_mean": (0.2, 2),
                    "friction_mean": (0.1, 1),
                    "seed_density": (1, 30),
                },
            },
            'FlowSettings': {
                # 'point_source' , 'line_source',
                # 'num_const_perm', 'num_var_perm_FD', 'num_var_perm'
                'flow_model': 'num_var_perm',
                # 'point_source' , 'line_source', 'num_const_perm'
                "permeability": -20,
                # 'point_source' , 'line_source', 'num_const_perm',
                # 'num_var_perm_FD', 'num_var_perm'
                "storage": -10,
                # 'point_source' , 'line_source', 'num_const_perm',
                # 'num_var_perm_FD', 'num_var_perm'
                "viscosity": 1e-3,
                # 'num_var_perm_FD','num_var_perm'
                'zero_stress_aperture': -4,
                # 'num_var_perm_FD','num_var_perm'
                'residual_aperture': -8,
                # 'num_var_perm_FD',#'num_var_perm'
                'frac_normal_stiffness': 11,
                'shearing_aperture_factor': 0,  # 'num_var_perm'
                'max_shear_frac_aperture': -3,  # 'num_var_perm'
                'shear_scaling_factor': -12,  # 'num_var_perm'
                'damping_parameter': 0.6,
                'threshold_pressure': 10,
                'fix_normal_stress': 21,
                'frac_spacing': 1,
                # default to 300, 'point_source' , 'line_source',
                # 'num_const_perm', 'num_var_perm'
                'max_grid_ext': 300,
                # default to 100, 'line_source',
                # 'num_const_perm', 'num_var_perm'
                'layer_thick': 100,
                # default to 94, 'point_source
                'cross_area': 94,
                'FlowBounds': {
                    # 'point_source' , 'line_source',
                    # 'num_const_perm', 'num_var_perm'
                    "storage": (-12, -11),
                    # 'point_source' , 'line_source', 'num_const_perm'
                    "permeability": (-20, -13),
                    'zero_stress_aperture': (-5, -3),  # 'num_var_perm'
                    'residual_aperture': (-10, -4),  # 'num_var_perm'
                    'shearing_aperture_factor': (-2, 2),  # 'num_var_perm'
                    'max_shear_frac_aperture': (-5, -2),  # 'num_var_perm'
                    'shear_scaling_factor': (-15, -10),  # 'num_var_perm'
                },
            }
        },
        'model_settings': {
            'well_section_id': str(section_id),
            # WGS84 coordinates of the injection point
            # Third dimension according to the seismicity
            # observation data in meters!
            "injection_point": reference,
            'local_proj_string': proj_string,
            'n_phases': 8,
            'epoch_duration': 600,
        }
    }
    # # # initialize the model

    # model_output = run_hm1d(data)  # unvalidated input
    model_output = run_hm1d_validated(data)  # validated input

    ############################################################
    # results are a list of ForecastGRRateGrids, they can for example
    # be accessed like this:
    print('\n\n'
          f'100 Simulations for timestep from {model_output[-1].starttime} '
          f'until {model_output[-1].endtime}\n',
          model_output[-1])

    # it is easy to add the datetime index to the dataframe:
    print('\n\n'
          'Same dataframe with the datetime index added:\n',
          model_output[-1].add_time_index())

    # as well as concatenating all the simulations into one dataframe:
    full_df = pd.concat([x.add_time_index() for x in model_output])
    print('\n\n'
          'All simulations concatenated into one dataframe:\n',
          full_df)

    ############################################################
    # you have also available here the:

    # input hydraulics (history) as a DataFrame
    print(hydraulics)

    # input injection plan as a DataFrame
    print(hydraulics_plan)

    # input seismicity as a seismostats Catalog
    print(catalog)

    return model_output


if __name__ == "__main__":
    result = main()
