import json
from datetime import datetime

import pandas as pd

from hm1d.hermes import run_hm1d_validated

with open("data/seismicityobservations.xml") as f:
    SEISMICITY_OBSERVATIONS = f.read()

with open("data/injectionobservations.json") as f:
    INJECTION_OBSERVATIONS = json.load(f)

with open("data/injectionplans.json") as f:
    INJECTION_PLANS = json.load(f)

POLYGON = \
    "POLYGON ((8.472517779144205 46.50848703012095, "\
    "8.472569275684041 46.51118547825671, "\
    "8.476477746198025 46.5111498555196, "\
    "8.47642605486892 46.50845140914097, "\
    "8.472517779144205 46.50848703012095))"


def run():
    data = {
        'forecast_start': datetime(2024, 12, 11, 8, 0, 0),
        'forecast_end': datetime(2024, 12, 12, 23, 0, 0),
        'injection_observation': INJECTION_OBSERVATIONS,
        'injection_plan': INJECTION_PLANS,
        'bounding_polygon': POLYGON,
        'depth_min': -150,
        'depth_max': 150,
        'seismicity_observation': SEISMICITY_OBSERVATIONS,
        "model_parameters": {
            "Delta_tau": 600,
            "max_iter_DE": 50,
            "DE_verbose": True,
            "SeisSettings": {
                "seis_model": "CAPS",
                "NSim": 50,
                "seed_model_version": 2,
                "seed_density": 5,
                "inter_sigma1": 26.5,
                "inter_sigma3": 16.7,
                "sigma1_std": 20,
                "sigma3_std": 12,
                "cohesion_mean": 1.0,
                "friction_mean": 0.6,
                "criticality": 0,
                "stressdrop_coeff": 0.1,
                "phydro": 4.0,
                "b_min": 1.0,
                "b_max": 5.0,
                "b_min_diff_stress": 50,
                "b_max_diff_stress": 0,
                "b_vs_depth_range": True,
                "CAPS_sigma_distr": 0.05,
                "retriggering": True,
                "dM": 0.05,
                "SeisBounds": {
                    "sigma1_std": [
                        0.01,
                        10
                    ],
                    "sigma3_std": [
                        0.01,
                        10
                    ],
                    "stressdrop_coeff": [
                        0.001,
                        0.01
                    ],
                    "cohesion_mean": [
                        0.2,
                        2
                    ],
                    "friction_mean": [
                        0.1,
                        1
                    ],
                    "seed_density": [
                        1,
                        30
                    ]
                }
            },
            "FlowSettings": {
                "flow_model": "line_source",
                "permeability": -20,
                "storage": -10,
                "viscosity": 0.001,
                "zero_stress_aperture": -4,
                "residual_aperture": -8,
                "frac_normal_stiffness": 11,
                "shearing_aperture_factor": 0,
                "max_shear_frac_aperture": -3,
                "shear_scaling_factor": -12,
                "damping_parameter": 0.6,
                "threshold_pressure": 10,
                "fix_normal_stress": 21,
                "frac_spacing": 1,
                "max_grid_ext": 300,
                "layer_thick": 2,
                "cross_area": 94,
                "FlowBounds": {
                    "storage": [
                        -12,
                        -9
                    ],
                    "permeability": [
                        -20,
                        -13
                    ],
                    "zero_stress_aperture": [
                        -5,
                        -3
                    ],
                    "residual_aperture": [
                        -10,
                        -4
                    ],
                    "shearing_aperture_factor": [
                        -2,
                        2
                    ],
                    "max_shear_frac_aperture": [
                        -5,
                        -2
                    ],
                    "shear_scaling_factor": [
                        -15,
                        -10
                    ]
                }
            }
        },
        "model_settings": {
            "well_section_id": "257b1702-37d4-4eff-b129-060c9c93c5b3",
            "epoch_duration": 600,
            "n_phases": 8,
            "injection_point": [
                8.471566153121655,
                46.512499740957466,
                -1469.236
            ],
            "local_proj_string": "epsg:2056"
        },
    }
    # # # initialize the model

    # model_output = run_hm1d(data)  # unvalidated input
    model_output = run_hm1d_validated(data)  # validated input

    ############################################################
    # results are a list of ForecastGRRateGrids, they can for example
    # be accessed like this:
    print('\n\n'
          f'100 Simulations for timestep from {model_output[-1].starttime} '
          f'until {model_output[-1].endtime}\n',
          model_output[-1])

    # it is easy to add the datetime index to the dataframe:
    print('\n\n'
          'Same dataframe with the datetime index added:\n',
          model_output[-1].add_time_index())

    # as well as concatenating all the simulations into one dataframe:
    full_df = pd.concat([x.add_time_index() for x in model_output])
    print('\n\n'
          'All simulations concatenated into one dataframe:\n',
          full_df)


if __name__ == '__main__':
    run()
